﻿//============================================================================
//DSS REST API Learning Tutorial 10: main program code
//Goal:  search for a list of instruments, and display their details.
//       This can serve to validate the list of instruments, if the input
//       and output identifier types (Ric, Cusip, Isin, etc.) are the same;
//       if that is the case, list the non validated ones.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Linq;
using DataScope.Select.Api.Content;
using DataScope.Select.Core.RestApi;
using DataScope.Select.Api.Search;
using DssRestfulApiTutorials;
using System.Collections.Generic;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //Expose the searchContext from dssClient:
            SearchContext searchContext = dssClient.searchContext;

            //-----------------------------------------------------------------
            //The instruments in the input list must all be of the same type; we
            //cannot mix several types (Ric, Cusip, Isin, etc.) in a single query.
            //Set the input and output identifier types to the same value if you
            //want to validate the input list.
            //Here we use Rics (comments include a variant for OrgIds):
            //-----------------------------------------------------------------
            //Rics:
            IdentifierType inputInstrumentsIdentifierType = IdentifierType.Ric;
            string[] inputInstruments = { "US10YT=RRPS", "JPY=", "IBM.N", "INVALID.RIC" };
            IdentifierType preferredResultIdentifierType = IdentifierType.Ric;
            //OrgIds:
            //IdentifierType inputInstrumentsIdentifierType = IdentifierType.OrgId;
            //string[] inputInstruments = { "18174", "18175" };
            //IdentifierType preferredResultIdentifierType = IdentifierType.OrgId;

            //-----------------------------------------------------------------
            //Let us assume they are in an array (maybe read from a database).
            //The API expects a comma separated string as input, so we transform it.
            //-----------------------------------------------------------------
            string inputInstrumentsString = string.Join(",", inputInstruments);
            Console.WriteLine("Input instruments list:\n {0}", inputInstrumentsString);

            //-----------------------------------------------------------------
            //Initialize variables that will serve for validation:
            //-----------------------------------------------------------------
            List<string> validatedInstruments = new List<string>();
            string[] nonValidatedInstruments = inputInstruments;

            //-----------------------------------------------------------------
            //Perform an instrument search.
            //The search criteria can be modified to suit your needs:
            //-----------------------------------------------------------------
            searchContext.Preferences.MaxPageSize = 100;
            IDssEnumerable<ValidatedInstrument> resultingInstruments =
                searchContext.InstrumentSearch(
                    AllInstrumentSearchRequest.Create(
                        new[] { 
                            InstrumentTypeGroup.CollatetizedMortgageObligations,
                            InstrumentTypeGroup.Commodities, InstrumentTypeGroup.Equities,
                            InstrumentTypeGroup.Funds, InstrumentTypeGroup.FuturesAndOptions,
                            InstrumentTypeGroup.GovCorp, InstrumentTypeGroup.MortgageBackedSecurities,
                            InstrumentTypeGroup.Money, InstrumentTypeGroup.Municipals
                        },
                        inputInstrumentsIdentifierType,
                        inputInstrumentsString,
                        preferredResultIdentifierType));

            //-----------------------------------------------------------------
            //Display the details of the instruments returned in the results:
            //-----------------------------------------------------------------
            Console.WriteLine("\nSearch results:");
            foreach (ValidatedInstrument instrument in resultingInstruments)
            {
                Console.WriteLine("{0}\n - IdentifierType: {1}\n - InstrumentType: {2}\n - Status: {3}\n - Description: {4}",
                    instrument.Identifier,
                    instrument.IdentifierType, instrument.InstrumentType,
                    instrument.Status, instrument.Description);
                //Validation code:
                if (inputInstruments.Contains(instrument.Identifier))
                {
                    if (!validatedInstruments.Contains(instrument.Identifier))                        
                    {
                        validatedInstruments.Add(instrument.Identifier);                        
                    }
                }
                nonValidatedInstruments = nonValidatedInstruments.Where(val => val != instrument.Identifier).ToArray();
            }
            int numberOfResults = resultingInstruments.Count();
            Console.WriteLine("\nNumber of results: {0}", numberOfResults);

            //-----------------------------------------------------------------
            //Check if all instruments were validated, list those that were not.
            //This only works if the identifier type of the input instruments is
            //the same as that of the results; we skip it otherwise:
            //-----------------------------------------------------------------
            if (inputInstrumentsIdentifierType == preferredResultIdentifierType)
            {
                Console.WriteLine("Number of validated instruments: {0}", validatedInstruments.Count);
                int numberOfNonValidatedInstruments = nonValidatedInstruments.Length;
                Console.WriteLine("Number of non validated instruments: {0}", numberOfNonValidatedInstruments);
                if (numberOfNonValidatedInstruments > 0)
                {
                    string nonValidatedInstrumentsString = string.Join(", ", nonValidatedInstruments);
                    Console.WriteLine("\nNon validated instruments list:\n {0}", nonValidatedInstrumentsString);
                }
            }

            DebugPrintAndWaitForEnter("\nSearch completed.");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

    }
}
