//============================================================================
//DSS REST API Learning Tutorial 2: main program code
//Goal: programmatic GUI control:
//      create an instrument list, a report template for EoD (End of Day) pricing,
//	    an immediate schedule that runs once,
//	    an EoD recurring schedule that waits for the error corrected data to be available;
//	    cleanup: delete them all when done.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Extractions.SubjectLists;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ReportTemplates;
using DataScope.Select.Api.Extractions.Schedules;
using DataScope.Select.Api;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";


        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //Expose the extractionsContext from dssClient:
            ExtractionsContext extractionsContext = dssClient.extractionsContext;

            //-----------------------------------------------------------------
            //Create and populate an instrument list:
            //-----------------------------------------------------------------
            //Create an array of instrument identifiers:
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = new[] 
            {
                new InstrumentIdentifier
                {
                    Identifier = "IBM.N",
                    IdentifierType = IdentifierType.Ric,
                    UserDefinedIdentifier = "EQUITYTEST"
                },
                new InstrumentIdentifier
                {
                    Identifier = "438516AC0",
                    IdentifierType = IdentifierType.Cusip,
                    UserDefinedIdentifier = "BONDTEST"
                }
            };

            //Create a new (empty) instrument list:
            var instrumentList = new InstrumentList { Name = "myInstrumentListName" };
            extractionsContext.InstrumentListOperations.Create(instrumentList);

            DebugPrintAndWaitForEnter("Returned instrument list ID: " +
                instrumentList.ListId + "\nCheck The DSS GUI...");

            //Populate the instrument list by appending the array of instrument identifiers:
            extractionsContext.InstrumentListOperations.AppendIdentifiers(
                instrumentList, instrumentIdentifiers, false);

            DebugPrintAndWaitForEnter("Instruments have been added to my instrument list." +
                "\nCheck The DSS GUI...");

            //-----------------------------------------------------------------
            //Create a report template:
            //-----------------------------------------------------------------
            //Define a report template, for EOD (End Of Day) pricing:
            EndOfDayPricingReportTemplate eodReportTemplate = new EndOfDayPricingReportTemplate();
            eodReportTemplate.Name = "myEodTemplateName";
            eodReportTemplate.OutputFormat = ReportOutputFormat.CommaSeparatedValues;

            //Add some field names to the report template:
            eodReportTemplate.ContentFields.AddRange("Instrument ID", "Security Description",
                "Universal Close Price Date", "Universal Close Price");

            //Create the report template, using the defined template:
            extractionsContext.ReportTemplateOperations.Create(eodReportTemplate);

            DebugPrintAndWaitForEnter("Returned EOD report template ID: " +
                eodReportTemplate.ReportTemplateId + "\nCheck The DSS GUI...");

            //-----------------------------------------------------------------
            //Create some schedules:
            //-----------------------------------------------------------------
            //Create a schedule for an immediate extraction:
            Schedule immediateSchedule = new Schedule
            {
                Name = "myImmediateSchedule",
                TimeZone = TimeZone.CurrentTimeZone.StandardName,
                Recurrence = ScheduleRecurrence.CreateSingleRecurrence(DateTimeOffset.UtcNow, true),
                Trigger = ScheduleTrigger.CreateImmediateTrigger(true),
                ListId = instrumentList.ListId,
                ReportTemplateId = eodReportTemplate.ReportTemplateId,
                OutputFileName = "myImmediateExtractionOutput.out",
            };

            extractionsContext.ScheduleOperations.Create(immediateSchedule);

            DebugPrintAndWaitForEnter("Returned immediate extraction schedule ID: " +
                immediateSchedule.ScheduleId + "\nCheck the DSS GUI...");

            //Create a recurring EOD schedule, with a trigger:
            Schedule eodSchedule = new Schedule
            {
                Name = "myEndOfBusinessDayCorrectedDataRecurringSchedule",
                TimeZone = "Eastern Standard Time",
                Recurrence = ScheduleRecurrence.CreateWeeklyRecurrence(
                    new[] { DayOfWeek.Monday, DayOfWeek.Tuesday, DayOfWeek.Wednesday,
                            DayOfWeek.Thursday, DayOfWeek.Friday }),
                //Wait for the latest end of day data before running the extraction:
                Trigger = ScheduleTrigger.CreateDataAvailabilityTrigger(
                    true,                                      //Limit to today's data
                    ExtendedTriggerInformation.CreateEndOfDay(
                        true,                                  // Error Corrected
                        false),                                // Delta
                    DataAvailabilityTriggerTimeout.Create(
                        HourMinute.Create(23, 15),             //Extract if data does not arrive by 11:15 PM ...
                        0)),                                   //... on the same day
                ListId = instrumentList.ListId,
                ReportTemplateId = eodReportTemplate.ReportTemplateId,
                OutputFileName = "myEndOfBusinessDayCorrectedDataRecurringExtractionOutput.out"
            };

            extractionsContext.ScheduleOperations.Create(eodSchedule);

            DebugPrintAndWaitForEnter("Returned recurring extraction schedule ID: " +
                eodSchedule.ScheduleId + "\nCheck the DSS GUI...");

            //-----------------------------------------------------------------
            //Clean Up:
            //-----------------------------------------------------------------
            //The purpose of this tutorial is to demonstrate manipulating DSS just like
            //through the DSS GUI, so we will stop here and cleanup what we created.

            //What about data extractions ?
            //Tutorials 3 and 4 shows extractions in a context similar to this tutorial.
            //Tutorials 5, 6 and 7 show a streamlined approach to data extractions.
            //Both approaches are valid, and delivered as examples of what can be done.

            //Delete instrument list, report template and schedules:
            extractionsContext.InstrumentListOperations.Delete(instrumentList.ListId);
            extractionsContext.ReportTemplateOperations.Delete(eodReportTemplate.ReportTemplateId);
            extractionsContext.ScheduleOperations.Delete(immediateSchedule.ScheduleId);
            extractionsContext.ScheduleOperations.Delete(eodSchedule.ScheduleId);

            DebugPrintAndWaitForEnter("Clean up is done...");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }
    }
}
