//============================================================================
//DSS REST API Refactored Tutorial 2: main program code
//Goal: programmatic GUI control:
//      create an instrument list, a report template for EoD (End of Day) pricing,
//	    an immediate schedule that runs once,
//	    an EoD recurring schedule that waits for the error corrected data to be available;
//	    cleanup: delete them all when done.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ReportTemplates;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";


        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //-----------------------------------------------------------------
            //Create and populate an instrument list:
            //-----------------------------------------------------------------
            //Create an array of instrument identifiers:
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = CreateInstrumentIdentifiers();

            //Create, and populate, an instrument list:
            string instrumentListId = dssClient.CreateAndPopulateInstrumentList(
                "myInstrumentListName",
                instrumentIdentifiers);
            DebugPrintAndWaitForEnter("Returned instrument list ID: " +
                instrumentListId + "\nCheck The DSS GUI...");

            //-----------------------------------------------------------------
            //Create a report template:
            //-----------------------------------------------------------------
            //Create an array of field names:
            string[] requestedFieldNames = CreateRequestedFieldNames();

            //Define a report template, for EOD (End Of Day) pricing, using the list of field names:
            string reportTemplateId = dssClient.CreateEodReportTemplate(
                "myEodTemplateName",
                ReportOutputFormat.CommaSeparatedValues,
                requestedFieldNames);
            DebugPrintAndWaitForEnter("Returned EOD report template ID: " +
                reportTemplateId + "\nCheck The DSS GUI...");

            //-----------------------------------------------------------------
            //Create some schedules:
            //-----------------------------------------------------------------
            //Create a schedule for an immediate extraction:
            string immediateScheduleId = dssClient.CreateImmediateSchedule(
                "myImmediateSchedule",
                instrumentListId,
                reportTemplateId,
                "myImmediateExtractionOutput.csv");
            DebugPrintAndWaitForEnter("Returned immediate extraction schedule ID: " +
				immediateScheduleId + "\n Check the DSS GUI...");

            //Create a recurring EOD schedule, with a trigger:
            string recurringScheduleId = dssClient.CreateEndOfBusinessDayCorrectedDataRecurringSchedule(
                "myEndOfBusinessDayCorrectedDataRecurringSchedule",
                instrumentListId,
                reportTemplateId,
                "myEndOfBusinessDayCorrectedDataRecurringExtractionOutput.csv");
            DebugPrintAndWaitForEnter("Returned recurring extraction schedule ID: " +
				recurringScheduleId + "\n Check the DSS GUI...");

            //-----------------------------------------------------------------
            //Clean Up:
            //-----------------------------------------------------------------
            //The purpose of this tutorial is to demonstrate manipulating DSS just like
            //through the DSS GUI, so we will stop here and cleanup what we created.

            //What about data extractions ?
            //Tutorials 3 and 4 shows extractions in a context similar to this tutorial.
            //Tutorials 5, 6 and 7 show a streamlined approach to data extractions.
            //Both approaches are valid, and delivered as examples of what can be done.

            //Delete instrument list, report template and schedules:
            dssClient.DeleteInstrumentList(instrumentListId);
            dssClient.DeleteReportTemplate(reportTemplateId);
            dssClient.DeleteSchedule(immediateScheduleId);
            dssClient.DeleteSchedule(recurringScheduleId);
            DebugPrintAndWaitForEnter("Clean up is done...");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        //Create an array of instrument identifiers:
        static IEnumerable<InstrumentIdentifier> CreateInstrumentIdentifiers()
        {
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = new[] 
            {
                new InstrumentIdentifier
                {
                    Identifier = "IBM.N",
                    IdentifierType = IdentifierType.Ric,
                    UserDefinedIdentifier = "EQUITYTEST"
                },
                new InstrumentIdentifier
                {
                    Identifier = "438516AC0",
                    IdentifierType = IdentifierType.Cusip,
                    UserDefinedIdentifier = "BONDTEST"
                }
            };
            return instrumentIdentifiers;
        }

        //Create an array of field names:
        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = { "Instrument ID", "Security Description",
                                             "Universal Close Price Date", "Universal Close Price" };
            return requestedFieldNames;
        }
    }
}
