//============================================================================
//DSS REST API Learning Tutorial 3: DssClient class
//DssClient class responsibilities:
//Connection:
//  Hold DSS server URI.
//  Connect (and authenticate) to the DSS REST server, by creating an extraction context.
//  Retrieve the session token from the extraction context.
//DSS GUI manipulation:
//  Create an instrument list.
//  Append identifiers to an instrument list.
//  Create an instrument list and populate it with identifiers.
//  Create an EOD (End Of Day) report template.
//  Create an immediate schedule.
//  Create an EOD corrected data recurring schedule.
//  Delete an instrument list.
//  Delete a report template.
//  Delete a schedule.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Extractions.SubjectLists;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ReportTemplates;
using DataScope.Select.Api.Extractions.Schedules;
using DataScope.Select.Api;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class DssClient
    {
        //extractionsContext temporarily set to public, to expose it to our learning program.
        public ExtractionsContext extractionsContext;

        private Uri dssUri = new Uri("https://selectapi.datascope.refinitiv.com/RestApi/v1/");
        


        public void ConnectToServer(string dssUserName, string dssUserPassword)
        {
            extractionsContext = new ExtractionsContext(dssUri, dssUserName, dssUserPassword);
        }

        public string SessionToken
        {
            //The session token is only generated if the server connection is successful.
            get { return extractionsContext.SessionToken; }
        }

        //Create a new (empty) instrument list:
        public InstrumentList CreateInstrumentList(string instrumentListName)
        {
            InstrumentList instrumentList = new InstrumentList { Name = instrumentListName };
            extractionsContext.InstrumentListOperations.Create(instrumentList);

            return instrumentList;
        }

        //Populate the instrument list by appending the array of instrument identifiers:
        public void AppendIdentifiersToInstrumentList(
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers, InstrumentList instrumentList)
        {
            extractionsContext.InstrumentListOperations.AppendIdentifiers(
                instrumentList, instrumentIdentifiers, false);
        }

        //Create, and populate, an instrument list, by appending the array of instrument identifiers:
        public string CreateAndPopulateInstrumentList(
            string instrumentListName, IEnumerable<InstrumentIdentifier> instrumentIdentifiers)
        {
            InstrumentList instrumentList = CreateInstrumentList(instrumentListName);
            AppendIdentifiersToInstrumentList(instrumentIdentifiers, instrumentList);
            return instrumentList.ListId;
        }

        //Define a report template, for EOD (End Of Day) pricing:
        public string CreateEodReportTemplate(
            string reportTemplateName, ReportOutputFormat reportOutputFormat, string[] contentFieldNames)
        {
            EndOfDayPricingReportTemplate reportTemplate = new EndOfDayPricingReportTemplate
            {
                Name = reportTemplateName,
                OutputFormat = reportOutputFormat
            };
            reportTemplate.ContentFields.AddRange(contentFieldNames);

            extractionsContext.ReportTemplateOperations.Create(reportTemplate);
            return reportTemplate.ReportTemplateId;
        }

        //Create a schedule for an immediate extraction:
        public string CreateImmediateSchedule(
            string scheduleName, string instrumentListId, string reportTemplateId, string outputFileName)
        {
            Schedule schedule = new Schedule
            {
                Name = scheduleName,
                TimeZone = TimeZone.CurrentTimeZone.StandardName,
                Recurrence = ScheduleRecurrence.CreateSingleRecurrence(DateTimeOffset.UtcNow, true),
                Trigger = ScheduleTrigger.CreateImmediateTrigger(true),
                ListId = instrumentListId,
                ReportTemplateId = reportTemplateId,
                OutputFileName = outputFileName
            };

            extractionsContext.ScheduleOperations.Create(schedule);
            return schedule.ScheduleId;
        }

        //Create a recurring EOD schedule, with a trigger:
        public string CreateEndOfBusinessDayCorrectedDataRecurringSchedule(
            string scheduleName, string instrumentListId, string reportTemplateId, string outputFileName)
        {
            Schedule schedule = new Schedule
            {
                Name = scheduleName,
                TimeZone = "Eastern Standard Time",
                Recurrence = ScheduleRecurrence.CreateWeeklyRecurrence(
                    new[] { DayOfWeek.Monday, DayOfWeek.Tuesday, DayOfWeek.Wednesday,
                            DayOfWeek.Thursday, DayOfWeek.Friday }),
                //Wait for the latest end of day data before running the extraction:
                Trigger = ScheduleTrigger.CreateDataAvailabilityTrigger(
                    true,                                      //Limit to today's data
                    ExtendedTriggerInformation.CreateEndOfDay(
                        true,                                  // Error Corrected
                        false),                                // Delta
                    DataAvailabilityTriggerTimeout.Create(
                        HourMinute.Create(2, 0),               //Time Out - wait 2 hours
                        0)),                                   //Trigger Timeout Day Offset - 0 days
                ListId = instrumentListId,
                ReportTemplateId = reportTemplateId,
                OutputFileName = outputFileName
            };

            extractionsContext.ScheduleOperations.Create(schedule);
            return schedule.ScheduleId;
        }

        //Delete instrument list:
        public void DeleteInstrumentList(string instrumentListId)
        {
            extractionsContext.InstrumentListOperations.Delete(instrumentListId);
        }

        //Delete report template:
        public void DeleteReportTemplate(string reportTemplateId)
        {
            extractionsContext.ReportTemplateOperations.Delete(reportTemplateId);
        }

        //Delete schedule:
        public void DeleteSchedule(string immediateScheduleId)
        {
            extractionsContext.ScheduleOperations.Delete(immediateScheduleId);
        }
    }
}
