//=============================================================================
//DSS REST API Refactored Tutorial 3: main program code
//Goals: create an instrument list and check the data validation results;
//       create a report template for EoD (End of Day) pricing,
//       and an EoD recurring schedule that runs at a specific time;
//       launch an extraction, wait for it to complete;
//       retrieve only the data file, display its contents;
//       clean up.
//This simple example is a variation of Tutorial 2.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.SubjectLists;
using DataScope.Select.Api.Extractions.ReportTemplates;
using DataScope.Select.Api.Extractions.Schedules;
using DataScope.Select.Api.Extractions.ReportExtractions;
using DataScope.Select.Api.Core;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";


        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //Expose the extractionsContext from dssClient:
            ExtractionsContext extractionsContext = dssClient.extractionsContext;

            //-----------------------------------------------------------------
            //Create and populate an instrument list:
            //-----------------------------------------------------------------
            //Create an array of instrument identifiers (of which some are invalid):
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = CreateInstrumentIdentifiers();

            //Create a new (empty) instrument list:
            InstrumentList instrumentList = dssClient.CreateInstrumentList("myInstrumentListName");
            string instrumentListId = instrumentList.ListId;
            Console.WriteLine("Returned instrument list ID: " + instrumentListId + "\n");

            //Populate the instrument list by appending the array of instrument identifiers,
            //capture the append result:
            InstrumentsAppendIdentifiersResult appendResult =
                dssClient.AppendIdentifiersToInstrumentList(
                    instrumentIdentifiers, instrumentList);
            int instrumentIdentifiersCount = instrumentIdentifiers.Count();
            Console.WriteLine("Attempted to add " + instrumentIdentifiersCount + " instruments to the list.\n");

            //-----------------------------------------------------------------
            //Display the append results:
            //  instrument identifiers validation,
            //  instrument identifiers segmentation by asset class.
            //-----------------------------------------------------------------
            //Display the instrument identifiers validation results, contained in the append result:
            DisplayAppendResults(appendResult);

            //-----------------------------------------------------------------
            //Create a report template:
            //-----------------------------------------------------------------
            //Create an array of field names:
            string[] requestedFieldNames = CreateRequestedFieldNames();

            //Define a report template, for EOD (End Of Day) pricing, using the list of field names:
            string reportTemplateId = dssClient.CreateEodReportTemplate(
                "myEodTemplateName",
                ReportOutputFormat.CommaSeparatedValues,
                requestedFieldNames);
            DebugPrintAndWaitForEnter("Returned EOD report template ID: " + reportTemplateId);

            //-----------------------------------------------------------------
            //Create a schedule:
            //-----------------------------------------------------------------
            //Create a daily recurring EOD schedule.
            //For testing purposes we set it 2 minutes in the future:
            int waitMinutes = 2;
            DateTimeOffset dateTimeNow = DateTimeOffset.Now;
            DateTimeOffset dateTimeSchedule = dateTimeNow.AddMinutes(waitMinutes);
            int scheduleHour = dateTimeSchedule.Hour;
            int scheduleMinute = dateTimeSchedule.Minute;
            
            string recurringScheduleId = dssClient.CreateTimedRecurringSchedule(
                "myRecurringSchedule", instrumentListId, reportTemplateId,
                scheduleHour, scheduleMinute, "myTimedRecurringExtractionOutput.csv");
            DebugPrintAndWaitForEnter("Returned recurring schedule ID: " + recurringScheduleId);

            //Retrieve the schedule by using its name, we need it in the next code block:
            Schedule recurringSchedule =
                extractionsContext.ScheduleOperations.GetByName("myRecurringSchedule");

            //-----------------------------------------------------------------
            //Launch the extraction, wait for it to complete:
            //-----------------------------------------------------------------
            try
            {
                Console.WriteLine("{0:T} We set a schedule for {1:t} ... ", DateTime.Now, dateTimeSchedule);
                Console.WriteLine("We now wait for the next extraction, please be patient ...");

                //Launch the extraction. It will execute at the set trigger time, i.e. in 2 minutes:
                //Wait for it to run, poll once a minute, time out after 5 minutes:
                TimeSpan timeOut = TimeSpan.FromMinutes(5);
                int pollPeriod = 60;
                ReportExtraction extraction =
                    extractionsContext.ScheduleOperations.WaitForNextExtraction(recurringSchedule, pollPeriod, timeOut);

                Console.WriteLine("\n{0:T} Extraction done.", DateTime.Now);

                //Display extraction status with details:
                Console.WriteLine("\nStatus: {0} - Detailed status: {1}\nExtraction start-end (UTC):\n{2} - {3}",
                    extraction.Status, extraction.DetailedStatus,
                    extraction.ExtractionStartUtc, extraction.ExtractionEndUtc);

                //Load the files list.
                //Notes:
                //- The toolkit only populates collections inside entities when you explicitly
                //  call LoadProperty(), so we do that to retrieve the files.
                extractionsContext.LoadProperty(extraction, "Files");
                Console.WriteLine("File count: " + extraction.Files.Count + "\n");

                //The extraction will generate 3 files:
                //  The RIC maintenance note.
                //      Contains information on name changes etc. This file could be empty.
                //  The extracted data itself.
                //      As this extraction is for EOD data, the file type is "Full",
                //      independent of preference settings.
                //      Partial deliveries will only occur for intraday extractions,
                //      that could be subject to embargoes, as illustrated in Tutorial 4.
                //  The extraction notes.
                //      Contains details of the extraction, including information on applied embargoes.

                //Retrieve only the extracted data file, ignore the RIC maintenance and
                //extraction notes files:
                ExtractedFile extractedDataFile =
                    extraction.Files.FirstOrDefault(f => f.FileType == ExtractedFileType.Full);

                if (extractedDataFile != null)
                {
                    Console.WriteLine("Extracted data file (ignore the others):");
                    DisplayFileDetailsAndContents(extractionsContext, extractedDataFile, 1);
                }
                else { DebugPrintAndWaitForEnter("ERROR: extracted data file was null"); }
            }
            catch (TimeoutException ex)
            {
                Console.WriteLine("\n{0:T} ERROR: extraction timed out:\n", DateTime.Now);
                DebugPrintAndWaitForEnter(ex.Message);
            }
            catch (Exception ex)
            {
                Console.WriteLine("\n{0:T} ERROR: extraction failed:\n", DateTime.Now);
                DebugPrintAndWaitForEnter(ex.Message);
            }

            //-----------------------------------------------------------------
            //Clean Up:
            //-----------------------------------------------------------------
            //Delete instrument list, report template and schedule:
            dssClient.Cleanup(dssClient, instrumentListId, reportTemplateId, recurringScheduleId);
            DebugPrintAndWaitForEnter("Clean up is done, will now exit ...");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        //Create an array of instrument identifiers:
        static IEnumerable<InstrumentIdentifier> CreateInstrumentIdentifiers()
        {
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = new[] 
            {
                new InstrumentIdentifier
                    { Identifier = "438516AC0", IdentifierType = IdentifierType.Cusip,
                      UserDefinedIdentifier = "HONEYWELL_BOND" },
                new InstrumentIdentifier
                    { Identifier = "IBM.N", IdentifierType = IdentifierType.Ric,
                      UserDefinedIdentifier = "IBM_STOCK" },
                new InstrumentIdentifier
                    { Identifier = "6502.T", IdentifierType = IdentifierType.Ric,
                      UserDefinedIdentifier = "TOSHIBA_STOCK" },
                new InstrumentIdentifier
                    { Identifier = "UBSG.S", IdentifierType = IdentifierType.Ric,
                      UserDefinedIdentifier = "UBS_STOCK" },
                new InstrumentIdentifier
                    { Identifier = "ALVG.DE", IdentifierType = IdentifierType.Ric,
                      UserDefinedIdentifier = "ALLIANZ_STOCK" },
                new InstrumentIdentifier
                    { Identifier = "INVALID", IdentifierType = IdentifierType.Cusip,
                      UserDefinedIdentifier = "INVALID INSTRUMENT_1" },
                new InstrumentIdentifier
                    { Identifier = "JUNK.JUNK", IdentifierType = IdentifierType.Ric,
                      UserDefinedIdentifier = "INVALID INSTRUMENT_2" }
            };
            return instrumentIdentifiers;
        }

        //Display the instrument identifiers validation results, contained in the append result:
        static void DisplayAppendResults(InstrumentsAppendIdentifiersResult appendResult)
        {
            Console.WriteLine("Instrument identifiers validation results:");
            Console.WriteLine("Validation messages:");
            foreach (InstrumentValidationMessage message in appendResult.ValidationResult.Messages)
                Console.WriteLine("- " + message.Message, message.Severity);

            DebugPrintAndWaitForEnter("Valid instrument count: " +
                appendResult.ValidationResult.ValidInstrumentCount);

            //The standard segments are separations of data by asset class.
            //Display them:
            Console.WriteLine("Asset classes count: " + appendResult.ValidationResult.StandardSegments.Count);
            Console.WriteLine("Asset classes codes and descriptions: ");
            for (int i = 0; i < appendResult.ValidationResult.StandardSegments.Count; i++)
            {
                Console.WriteLine("- Code: " + appendResult.ValidationResult.StandardSegments[i].Code +
                    " - Description: " + appendResult.ValidationResult.StandardSegments[i].Description);
            }
            DebugPrintAndWaitForEnter("");
        }

        //Create an array of field names:
        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = {
                "Instrument ID", "Security Description",
                "Ask Price", "Ask Yield", "Asset Swap Spread",
                "Universal Close Price Date", "Universal Close Price",
                "Trade Date" };
            return requestedFieldNames;
        }

        static void DisplayFileDetailsAndContents(
            ExtractionsContext extractionsContext, ExtractedFile file, int fileNumber)
        {
            Console.WriteLine("================================ FILE DETAILS =================================" +
                "\nFile number: " + fileNumber +
                "\nFile name: " + file.ExtractedFileName +
                "\nFile type: " + file.FileType +
                "\nReceived date utc:" + file.ReceivedDateUtc +
                "\nFile size:" + file.Size);

            //Don't fetch the content if the file is empty:
            if (file.Size == 0) return;

            Console.WriteLine("================================ FILE CONTENTS ================================");
            DssStreamResponse streamResponse =
                extractionsContext.ExtractedFileOperations.GetReadStream(file);
            string streamContents;
            using (StreamReader reader = new StreamReader(streamResponse.Stream))
                Console.WriteLine(streamContents = reader.ReadToEnd());
            DebugPrintAndWaitForEnter("===============================================================================");
        }
    }
}
