//============================================================================
//DSS REST API Refactored Tutorial 4: main program code
//Goal: handle embargoes:
//      create an instrument list, a report template for intraday pricing,
//      and a schedule that runs once at a specific time;
//      wait for the schedule to launch an extraction, wait for it to complete;
//      handle partial / late deliveries that arise if it is under embargo;
//      retrieve and display all files (data and extraction notes);
//      clean up.
//============================================================================
//An embargo is a temporary block on the extraction of requested data.
//Embargoes can occur when requesting very recent data, for which the user is not
//permissioned. The DSS REST API checks the account for real time permissions for
//all exchanges in the request. Data for which there is no real time entitlement
//is embargoed.
//The embargo duration depends on the exchange, and can also vary during the day:
//- It can vary in length throughout the trading day.
//- An embargo does not apply outside the trading hours of an exchange.
//
//How does the DSS REST API handle embargoes ? It waits for the embargo to be
//lifted, and only extracts the data at the end of the embargo period.
//If there is a mix of embargoed and not embargoed data, behaviour can be tuned in
//the DSS general preferences settings, using the DSS GUI. One can choose between:
//- Receiving the entire set of requested data at the end of the longest embargo
//  period, in one single file.
//- Receiving several partial deliveries: first the non embargoed data, then the
//  embargoed data (either in one single file, or one file per embargo period).
//Notes:
//- Even if no embargoes apply, an extraction can take several minutes.
//  This depends on: requested data amount, data origin, DSS server load.
//- This sample is for scheduled extractions. In the future, partial extractions
//  will also be supported for on demand extractions.
//
//This example, a variation of Tutorials 2 and 3, shows how to handle embargoes.
//We use exchange Xetra (Germany) which has a 15 minutes embargo during market open.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.IO;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ReportTemplates;
using DataScope.Select.Api.Extractions.ReportExtractions;
using DataScope.Select.Api.Core;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //Expose the extractionsContext from dssClient:
            ExtractionsContext extractionsContext = dssClient.extractionsContext;

            //-----------------------------------------------------------------
            //Create and populate an instrument list:
            //-----------------------------------------------------------------
            //Create an array of instrument identifiers:
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = CreateInstrumentIdentifiers();

            //Create, and populate, an instrument list:
            string instrumentListId = dssClient.CreateAndPopulateInstrumentList(
                "myEmbargoInstrumentListName",
                instrumentIdentifiers);
            DebugPrintAndWaitForEnter("Returned instrument list ID: " + instrumentListId);
            //Depending on your DSS data permissions, an embargo might be applied or not.
            //You can edit the method to add/remove instruments subject to embargo.

            //-----------------------------------------------------------------
            //Create a report template:
            //-----------------------------------------------------------------
            //Create an array of field names:
            string[] requestedFieldNames = CreateRequestedFieldNames();

            //Define a report template, for Intraday Pricing, using the list of field names:
            string reportTemplateId = dssClient.CreateIntradayPricingReportTemplate(
                "myEmbargoIntradayPricingTemplateName",
                ReportOutputFormat.CommaSeparatedValues,
                requestedFieldNames);
            DebugPrintAndWaitForEnter("Returned Intraday Pricing report template ID: " + reportTemplateId);
            //As it is an intraday request, the pricing fields could cause an embargo !

            //-----------------------------------------------------------------
            //Create a schedule:
            //-----------------------------------------------------------------
            //As an example create a schedule for today, that will run in 2 minutes
            //(it could also be an immediate schedule):
            int waitMinutes = 2;
            DateTimeOffset dateTimeNow = DateTimeOffset.Now;
            DateTimeOffset dateTimeSchedule = dateTimeNow.AddMinutes(waitMinutes);
            int scheduleHour = dateTimeSchedule.Hour;
            int scheduleMinute = dateTimeSchedule.Minute;

            string timedScheduleId = dssClient.CreateTimedSchedule(
                "myTimedSchedule",
                instrumentListId,
                reportTemplateId,
                dateTimeNow, scheduleHour, scheduleMinute,
                "myTimedExtractionOutput.csv");
            DebugPrintAndWaitForEnter("Returned timed extraction schedule ID: " + timedScheduleId);

            //-----------------------------------------------------------------
            //Wait for the DSS server to initiate the schedule's extraction:
            //-----------------------------------------------------------------
            Console.WriteLine("{0:T} We set a schedule for {1:t} ... ", DateTime.Now, dateTimeSchedule);

            //Wait for DSS servers to initiate schedule's extraction.
            //We may have to wait, depending on the DSS servers load:
            ReportExtraction extraction =
                dssClient.WaitForExtractionInitiation(extractionsContext, timedScheduleId);
            string extractionReportId = extraction.ReportExtractionId;

            Console.WriteLine("DSS server has initiated the scheduled extraction.\n");

            Console.WriteLine("{0:T} Initial extraction status:", DateTime.Now);
            Console.WriteLine("Report id: " + extractionReportId +
                "\nStatus: " + extraction.Status +
                " - Detailed status: " + extraction.DetailedStatus);
            //Expected extraction status values, occuring in time from left to right:
            //Status:          Pending,                                Processing,            Completed
            //Detailed status: Queued, WaitingForPreExtractCompletion, Processing, Embargoed, Done

            //-----------------------------------------------------------------
            //Wait for the extraction to complete, retrieve files:
            //-----------------------------------------------------------------
            //This extraction might result in an embargo, depending on our user
            //account permissions for the requested data, and the time of day.

            //Create a list to store the files that were retrieved:
            List<ExtractedFile> retrievedFiles = new List<ExtractedFile>();

            retrievedFiles = RetrieveExtractionFiles(extractionsContext, extraction);

            //-----------------------------------------------------------------
            //Display the extracted files:
            //-----------------------------------------------------------------
            //This tutorial displays all files once the extraction has completed.
            //Depending on the use case partial files could be treated immediately after reception,
            //as they arrive, in the RetrieveExtractionFiles method (see comment in method).
            DebugPrintAndWaitForEnter("\nNow we proceed to retrieve all the files and their content:");

            //The extraction will generate at least 3 files, maybe more:
            //  The RIC maintenance note.
            //      Contains information on name changes etc. This file could be empty.
            //  The extracted data itself.
            //      This can be 1 or more partial files, or a full file, depending on preferences.
            //      Later below we display the FileType, which tells us if it is partial or full.
            //  The extraction notes.
            //      Contains details of the extraction, including information on applied embargoes.
            int fileNumber = 0;
            foreach (ExtractedFile retrievedFile in retrievedFiles)
            {
                fileNumber++;
                DisplayFileDetailsAndContents(extractionsContext, retrievedFile, fileNumber);
            }

            //-----------------------------------------------------------------
            //Clean Up:
            //-----------------------------------------------------------------
            //Delete instrument list, report template and schedule:
            dssClient.Cleanup(dssClient, instrumentListId, reportTemplateId, timedScheduleId);
            DebugPrintAndWaitForEnter("Clean up is done, will now exit ...");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        //Create an array of instrument identifiers:
        static IEnumerable<InstrumentIdentifier> CreateInstrumentIdentifiers()
        {
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers = new[] 
            {
                new InstrumentIdentifier
                    { Identifier = "MX.TO", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "IBM.N", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "CSCO.O", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "0001.HK", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "6502.T", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "047810.KS", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "CARR.PA", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "UBSG.VX", IdentifierType = IdentifierType.Ric },
                new InstrumentIdentifier
                    { Identifier = "VOD.L", IdentifierType = IdentifierType.Ric },
                //new InstrumentIdentifier
                //    { Identifier = "ALVG.DE", IdentifierType = IdentifierType.Ric }
            };
            return instrumentIdentifiers;
        }

        //Create an array of field names:
        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = {
                "Instrument ID", "Security Description", "ISIN", "Asset ID",
                "Ask Price", "Ask Size", "Ask Yield", "Asset Swap Spread",
                "Asset Category", "Asset Category Description",
                "Asset Status", "Asset Status Description",
                "Asset Type", "Asset Type Description", "Asset SubType", "Asset SubType Description" };
            return requestedFieldNames;
        }

        //Wait for the extraction to complete, retrieve files:
        static List<ExtractedFile> RetrieveExtractionFiles(ExtractionsContext extractionsContext, ReportExtraction extraction)
        {
            string extractionReportId = extraction.ReportExtractionId;

            //Create a list to store the files that were retrieved:
            List<ExtractedFile> retrievedFiles = new List<ExtractedFile>();
            //Create a hash set to store the names of the files that were retrieved:
            HashSet<string> retrievedFileNames = new HashSet<string>();

            while (true)
            {
                //(Re-)load the files list, some may have arrived since the last loop iteration.
                //Note:
                //- The toolkit only populates collections inside entities when you explicitly call
                //  LoadProperty(), so we do that now to populate the files in extraction.Files.
                extractionsContext.LoadProperty(extraction, "Files");
                Console.WriteLine("Retrieved the updated file list. Extraction files count: " +
                    extraction.Files.Count);

                //Generate a list of new files, fitering out already retrieved ones:
                IEnumerable<ExtractedFile> newFiles =
                    extraction.Files.Where(f => !retrievedFileNames.Contains(f.ExtractedFileName));
                /*
                //The following also works, without requiring retrievedFileNames:
                IEnumerable<ExtractedFile> newFiles =
                    extraction.Files.Where(newfile =>
                        !retrievedFiles.Any(retfile =>
                            newfile.ExtractedFileName == retfile.ExtractedFileName));
                */

                //Add the new files to the list of retrieved files:
                foreach (ExtractedFile newFile in newFiles)
                {
                    retrievedFiles.Add(newFile);
                    retrievedFileNames.Add(newFile.ExtractedFileName);
                    Console.WriteLine("New file {0} name: {1}", retrievedFiles.Count, newFile.ExtractedFileName);
                    //If we want to treat new files immediately, we can do it here.
                    //Uncomment the next line to see that in action:
                    //DisplayFileDetailsAndContents(extractionsContext, newFile, retrievedFiles.Count);
                }

                //Exit loop if the extraction is completed:
                if (extraction.Status == ReportExtractionStatus.Completed)
                    break;

                Thread.Sleep(60 * 1000);  //Wait 1 minute

                //Retrieve this extractions updated report (by referring to its Id):
                extraction = extractionsContext.ReportExtractionOperations.Get(extractionReportId);
                Console.WriteLine("\n{0:T} Retrieved the updated extraction report:", DateTime.Now);
                Console.WriteLine("Status: " + extraction.Status +
                    " - Detailed status: " + extraction.DetailedStatus);
                //Extraction properties are emptied when retrieving the extractions updated
                //report, hence the need to call LoadProperty() to refresh the files list,
                //at the start of the loop.

            }  //End of while loop
            return retrievedFiles;
        }

        static void DisplayFileDetailsAndContents(
            ExtractionsContext extractionsContext, ExtractedFile file, int fileNumber)
        {
            Console.WriteLine("================================ FILE DETAILS =================================" +
                "\nFile number: " + fileNumber +
                "\nFile name: " + file.ExtractedFileName +
                "\nFile type: " + file.FileType +
                "\nReceived date utc:" + file.ReceivedDateUtc +
                "\nFile size:" + file.Size);
            if (file.FileType.ToString() == "Full")
            {
                Console.WriteLine("=================================== COMMENT ===================================");
                Console.WriteLine("File type is \"Full\": it contains all the data.");
            }
            if (file.FileType.ToString() == "Partial")
            {
                Console.WriteLine("=================================== COMMENT ===================================");
                Console.WriteLine("File type is \"Partial\":\n" +
                    "In absence of embargo this is the only data file, even if it is \"Partial\".\n" +
                    "If embargo(es) apply there are 2 data files or more.");
            }
            DebugPrintAndWaitForEnter("===============================================================================");

            //Don't fetch the content if the file is empty:
            if (file.Size == 0) return;

            Console.WriteLine("================================ FILE CONTENTS ================================");
            DssStreamResponse streamResponse =
                extractionsContext.ExtractedFileOperations.GetReadStream(file);
            string streamContents;
            using (StreamReader reader = new StreamReader(streamResponse.Stream))
                Console.WriteLine(streamContents = reader.ReadToEnd());
            if (streamContents.IndexOf("embargoed") != -1)
            {
                Console.WriteLine("=================================== COMMENT ===================================");
                Console.WriteLine("There was an embargo !");
            }
            DebugPrintAndWaitForEnter("===============================================================================");
        }
    }
}
