﻿//============================================================================
//DSS REST API Learning Tutorial 5: DssClient class
//DssClient class responsibilities:
//Connection:
//  Hold DSS server URI.
//  Connect (and authenticate) to the DSS REST server, by creating an extraction context.
//  Retrieve the session token from the extraction context.
//DSS GUI manipulation:
//  Create an instrument list.
//  Append identifiers to an instrument list (with validation result).
//  Create an instrument list and populate it with identifiers.
//  Create an EOD (End Of Day) report template.
//  Create an Intraday Pricing report template.
//  Create an immediate schedule.
//  Create a timed schedule.
//  Create a timed recurring schedule.
//  Create an EOD corrected data recurring schedule.
//  Delete an instrument list.
//  Delete a report template.
//  Delete a schedule.
//  Cleanup: delete a set of instrument list, report template and schedule.
//  Wait for DSS servers to initiate a schedule's extraction.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Extractions.SubjectLists;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ReportTemplates;
using DataScope.Select.Api.Extractions.Schedules;
using DataScope.Select.Api;
using DataScope.Select.Api.Extractions.ReportExtractions;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class DssClient
    {
        //extractionsContext temporarily set to public, to expose it to our learning program.
        public ExtractionsContext extractionsContext;

        private Uri dssUri = new Uri("https://selectapi.datascope.refinitiv.com/RestApi/v1/");
        

        public void ConnectToServer(string dssUserName, string dssUserPassword)
        {
            extractionsContext = new ExtractionsContext(dssUri, dssUserName, dssUserPassword);
        }

        public string SessionToken
        {
            //The session token is only generated if the server connection is successful.
            get { return extractionsContext.SessionToken; }
        }

        public InstrumentList CreateInstrumentList(string instrumentListName)
        {
            InstrumentList instrumentList = new InstrumentList { Name = instrumentListName };
            extractionsContext.InstrumentListOperations.Create(instrumentList);

            return instrumentList;
        }

        public InstrumentsAppendIdentifiersResult AppendIdentifiersToInstrumentList(
            IEnumerable<InstrumentIdentifier> instrumentIdentifiers, InstrumentList instrumentList)
        {
            InstrumentsAppendIdentifiersResult appendResult =
                extractionsContext.InstrumentListOperations.AppendIdentifiers(
                    instrumentList, instrumentIdentifiers, false);
            return appendResult;
        }

        public string CreateAndPopulateInstrumentList(
            string instrumentListName, IEnumerable<InstrumentIdentifier> instrumentIdentifiers)
        {
            InstrumentList instrumentList = CreateInstrumentList(instrumentListName);
            AppendIdentifiersToInstrumentList(instrumentIdentifiers, instrumentList);
            return instrumentList.ListId;
        }

        public string CreateEodReportTemplate(
            string reportTemplateName, ReportOutputFormat reportOutputFormat, string[] contentFieldNames)
        {
            EndOfDayPricingReportTemplate reportTemplate = new EndOfDayPricingReportTemplate
            {
                Name = reportTemplateName,
                OutputFormat = reportOutputFormat
            };
            reportTemplate.ContentFields.AddRange(contentFieldNames);

            extractionsContext.ReportTemplateOperations.Create(reportTemplate);
            return reportTemplate.ReportTemplateId;
        }

        //Define a report template, for Intraday Pricing:
        public string CreateIntradayPricingReportTemplate(
            string reportTemplateName, ReportOutputFormat reportOutputFormat, string[] contentFieldNames)
        {
            IntradayPricingReportTemplate reportTemplate = new IntradayPricingReportTemplate
            {
                Name = reportTemplateName,
                OutputFormat = reportOutputFormat,
                CompressionType = ReportCompressionType.None,
                Delimiter = ReportDelimiter.Pipe,
                DeliveryType = ReportDeliveryType.None
            };
            reportTemplate.ContentFields.AddRange(contentFieldNames);

            extractionsContext.ReportTemplateOperations.Create(reportTemplate);
            return reportTemplate.ReportTemplateId;
        }

        public string CreateImmediateSchedule(
            string scheduleName, string instrumentListId, string reportTemplateId, string outputFileName)
        {
            Schedule schedule = new Schedule
            {
                Name = scheduleName,
                TimeZone = TimeZone.CurrentTimeZone.StandardName,
                Recurrence = ScheduleRecurrence.CreateSingleRecurrence(DateTimeOffset.UtcNow, true),
                Trigger = ScheduleTrigger.CreateImmediateTrigger(true),
                ListId = instrumentListId,
                ReportTemplateId = reportTemplateId,
                OutputFileName = outputFileName
            };

            extractionsContext.ScheduleOperations.Create(schedule);
            return schedule.ScheduleId;
        }

        //Create a timed schedule:
        public string CreateTimedSchedule(
            string scheduleName, string instrumentListId, string reportTemplateId,
            DateTimeOffset scheduleDate, int scheduleHour, int scheduleMinute, string outputFileName)
        {
            Schedule schedule = new Schedule
            {
                Name = scheduleName,
                TimeZone = TimeZone.CurrentTimeZone.StandardName,
                Recurrence = ScheduleRecurrence.CreateSingleRecurrence(scheduleDate, isImmediate: false),
                Trigger = new TimeTrigger
                {
                    LimitReportToTodaysData = false,
                    At = new[] { new HourMinute { Hour = scheduleHour, Minute = scheduleMinute } }
                },
                ListId = instrumentListId,
                ReportTemplateId = reportTemplateId,
                OutputFileName = outputFileName
            };

            extractionsContext.ScheduleOperations.Create(schedule);
            return schedule.ScheduleId;
        }

        public string CreateTimedRecurringSchedule(
        string scheduleName, string instrumentListId, string reportTemplateId,
        int scheduleHour, int scheduleMinute, string outputFileName)
        {
            Schedule schedule = new Schedule
            {
                Name = scheduleName,
                TimeZone = TimeZone.CurrentTimeZone.StandardName,
                Recurrence = ScheduleRecurrence.CreateWeeklyRecurrence(
                    new[] { DayOfWeek.Monday, DayOfWeek.Tuesday, DayOfWeek.Wednesday,
                            DayOfWeek.Thursday, DayOfWeek.Friday }),
                //This trigger is on a set time (regardless of data availability):
                Trigger = ScheduleTrigger.CreateTimeTrigger(
                    false,                                   //Limit to today's data: no
                    new[] { new HourMinute { Hour = scheduleHour, Minute = scheduleMinute } }),
                ListId = instrumentListId,
                ReportTemplateId = reportTemplateId,
                OutputFileName = outputFileName
            };

            extractionsContext.ScheduleOperations.Create(schedule);
            return schedule.ScheduleId;
        }

        public string CreateEndOfBusinessDayCorrectedDataRecurringSchedule(
            string scheduleName, string instrumentListId, string reportTemplateId, string outputFileName)
        {
            Schedule schedule = new Schedule
            {
                Name = scheduleName,
                TimeZone = "Eastern Standard Time",
                Recurrence = ScheduleRecurrence.CreateWeeklyRecurrence(
                    new[] { DayOfWeek.Monday, DayOfWeek.Tuesday, DayOfWeek.Wednesday,
                            DayOfWeek.Thursday, DayOfWeek.Friday }),
                //Wait for the latest end of day data before running the extraction:
                Trigger = ScheduleTrigger.CreateDataAvailabilityTrigger(
                    true,                                      //Limit to today's data
                    ExtendedTriggerInformation.CreateEndOfDay(
                        true,                                  // Error Corrected
                        false),                                // Delta
                    DataAvailabilityTriggerTimeout.Create(
                        HourMinute.Create(2, 0),               //Time Out - wait 2 hours
                        0)),                                   //Trigger Timeout Day Offset - 0 days
                ListId = instrumentListId,
                ReportTemplateId = reportTemplateId,
                OutputFileName = outputFileName
            };

            extractionsContext.ScheduleOperations.Create(schedule);
            return schedule.ScheduleId;
        }

        public void DeleteInstrumentList(string instrumentListId)
        {
            extractionsContext.InstrumentListOperations.Delete(instrumentListId);
        }

        public void DeleteReportTemplate(string reportTemplateId)
        {
            extractionsContext.ReportTemplateOperations.Delete(reportTemplateId);
        }

        public void DeleteSchedule(string immediateScheduleId)
        {
            extractionsContext.ScheduleOperations.Delete(immediateScheduleId);
        }

        //Delete instrument list, report template and schedule:
        public void Cleanup(
            DssClient dssClient, string instrumentListId, string reportTemplateId, string scheduleId)
        {
            try { dssClient.DeleteInstrumentList(instrumentListId); }  catch { };
            try { dssClient.DeleteReportTemplate(reportTemplateId); }  catch { };
            try { dssClient.DeleteSchedule(scheduleId); }              catch { };
        }

        //Wait for DSS servers to initiate a schedule's extraction.
        //We may have to wait, depending on the DSS servers load:
        public ReportExtraction WaitForExtractionInitiation(
            ExtractionsContext extractionsContext, string scheduleId)
        {
            ReportExtraction extraction = null;
            while (extraction == null)
            {
                extraction =
                    extractionsContext.ReportExtractionOperations.GetByScheduleId(scheduleId).FirstOrDefault();
                if (extraction == null)
                {
                    Console.WriteLine("{0:T} Waiting for server to initiate scheduled extraction ... ", DateTime.Now);
                    Thread.Sleep(60 * 1000);  //Wait 1 minute
                }
            }
            return extraction;
        }
    }
}
