﻿//============================================================================
//DSS REST API Learning Tutorial 5: main program code
//Goal: move away from GUI control actions to directly interface with the API:
//      instrument lists: create an instrument list array, outside of DSS;
//      field lists: create a field list array, outside of DSS;
//      create and run an EoD on demand extraction, for JSON formatted data;
//      wait for it to complete;
//      display received data and extraction notes on screen.
//      Cleanup is not required in this case.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ExtractionRequests;
using DataScope.Select.Core.RestApi;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //Expose the extractionsContext from dssClient:
            ExtractionsContext extractionsContext = dssClient.extractionsContext;

            //-----------------------------------------------------------------
            //Now we move away from controlling the GUI via the API,
            //to directly interfacing with the API.
            //This is why, contrary to the previous tutorials, we do not create
            //an instrument list and report template on the DSS server.
            //-----------------------------------------------------------------
            
            //-----------------------------------------------------------------
            //Create an array outside of DSS, populate it with instruments:
            //-----------------------------------------------------------------
            //Instead of creating an instrument list in DSS, we create an array
            //of instrument identifiers that we will use in the extraction.
            //In the next tutorial we will use such an array to read a large
            //instrument list from file.

            //Create an array for the instrument identifiers, of size 5:
            InstrumentIdentifier[] instrumentIdentifiers = new InstrumentIdentifier[5];

            //In this example we populate it manually:
            instrumentIdentifiers[0] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Ric,
                Identifier = "IBM.N",
                UserDefinedIdentifier = "EQUITYTEST"
            };
            instrumentIdentifiers[1] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Cusip,
                Identifier = "438516AC0",
                UserDefinedIdentifier = "BONDTEST"
            };
            //The next 2 do not add the optional field, user defined identifier:
            instrumentIdentifiers[2] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Ric,
                Identifier = "0001.HK"                
            };
            instrumentIdentifiers[3] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Ric,
                Identifier = "VOD.L"
            };
            //The next one is a non existing identifier code, it will not return any data:
            instrumentIdentifiers[4] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Ric,
                Identifier = "DUMMY.DUMMY"
            };

            //-----------------------------------------------------------------
            //Create an array of field names:
            //-----------------------------------------------------------------
            //We do not create a report template, we only create an array of field names.
            //We will use the array in the extractions.
            string[] requestedFieldNames = CreateRequestedFieldNames();

            //-----------------------------------------------------------------
            //Create and run an on demand extraction. Retrieve JSON formatted data:
            //-----------------------------------------------------------------
            //As an example we choose to do an EOD (End Of Day) extraction:
            DebugPrintAndWaitForEnter("We created an array with 5 instrument identifiers, " +
                "and an array of field names, outside of DSS." +
                "\n\nNext we will launch a direct on demand EOD pricing extraction," +
                "\nusing these 2 arrays.");

            Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
            //Create an on demand extraction, of type EndOfDayPricingExtractionRequest,
            //using our instrument identifiers and requested field names arrays:
            EndOfDayPricingExtractionRequest extractionEod = new EndOfDayPricingExtractionRequest
            {
                IdentifierList = InstrumentIdentifierList.Create(instrumentIdentifiers),
                ContentFieldNames = requestedFieldNames
            };

            //Run the extraction.
            //This call is blocking, it returns when the extraction is completed:
            ExtractionResult extractionResult = extractionsContext.ExtractWithNotes(extractionEod);
            DssCollection<ExtractionRow> extractionDataRows = extractionResult.Contents;

            //-----------------------------------------------------------------
            //Data treatment starts here. We just display it on screen:
            //-----------------------------------------------------------------
            //Error handling:
            if (!extractionDataRows.Any())
                { DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?"); }
            else
            {
                //Extract the names of the data fields, from the first row of returned data:
                StringBuilder returnedFieldNames = new StringBuilder();
                ExtractionRow firstRow = extractionDataRows.First();
                foreach (string fieldName in firstRow.DynamicProperties.Keys)
                {
                    returnedFieldNames.Append(fieldName);
                    returnedFieldNames.Append(",");
                }
                DebugPrintAndWaitForEnter("Returned list of field names:\n" + returnedFieldNames.ToString());

                //Extract the actual data values, from all rows of returned data:
                Console.WriteLine("Returned field values:");
                Console.WriteLine("======================");
                string rowValuesInString = string.Empty;
                int validDataRows = 0;
                foreach (ExtractionRow row in extractionDataRows)
                {
                    IEnumerable<object> rowValues = row.DynamicProperties.Select(dp => dp.Value);
                    rowValuesInString = String.Join(", ", rowValues);
                    if (rowValuesInString != string.Empty)
                    {
                        Console.WriteLine(rowValuesInString);
                        validDataRows++;
                    }
                }

                //The number of data rows could be less than the number of instruments,
                //if data is not available for some instruments:
                DebugPrintAndWaitForEnter("\nExtraction completed." +
                    "\nNumber of data rows: " + extractionDataRows.Count() +
                    "\nNumber of valid (non empty) data rows: " + validDataRows);
            }

            //Display the extraction notes:
            if (extractionResult.Notes.Any())
            {
                Console.WriteLine("Extraction Notes:");
                Console.WriteLine("=================");
                foreach (String note in extractionResult.Notes)
                    Console.WriteLine(note);
            }
            else Console.WriteLine("Error: no extraction notes returned");
            DebugPrintAndWaitForEnter("");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = { "Instrument ID", "Security Description",
                                             "Universal Close Price Date", "Universal Close Price" };
            return requestedFieldNames;
        }
    }
}
