//============================================================================
//DSS REST API Learning Tutorial 6: main program code
//Goal:  start productizing our code:
//       instrument lists: create a large instrument list array outside of DSS,
//                         populated from an input file, with error handling;
//       field lists: create a field list array, outside of DSS;
//       We show 2 variants:
//       1) create and run an EoD on demand extraction, for JSON formatted data;
//       wait for it to complete;
//       display received data on screen and write it to a CSV file;
//       display received extraction notes on screen and write them to file;
//       analyze the extraction notes;
//       include some error checking.
//       2) repeat for GZipped CSV formatted data (use case: large data sets);
//       in this scenario we do not display the data on screen,
//       we just save it to a Gzipped CSV file;
//       Note) Cleanup is not required in this case.
//Limit: this code does not take into account the maximum size of DSS
//       instrument lists (75000 instruments). See Tutorial 7 for this.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using DataScope.Select.Api.Core;
using DataScope.Select.Api.Extractions;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ExtractionRequests;
using DataScope.Select.Core.RestApi;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        private static string outputDirectory = "C:\\DSS REST API\\";
        private static string dataOutputFile = outputDirectory + "DSS_API_tutorial_6l_output.csv";
        private static string notesOutputFile = outputDirectory + "DSS_API_tutorial_6l_notes.txt";
        private static string errorOutputFile = outputDirectory + "DSS_API_tutorial_6l_errors.txt";
        private static string gzipDataOutputFile = outputDirectory + "DSS_API_tutorial_6l_gzipOutput.csv.gz";
        private static string gzipNotesOutputFile = outputDirectory + "DSS_API_tutorial_6l_gzipNotes.txt";

        private static string inputDirectory = outputDirectory;
        private static string instrumentIdentifiersInputFile = inputDirectory + "DSS_API_10_input_file.csv";
        //File format: 1 instrument per line.
        //Line format: <identifier type>, <code>, <optional comment>
        //Example lines, first two are for a Cusip (CSP) and a RIC:
        //CSP,31352JRV1
        //RIC,ALVG.DE
        //CHR,0#.FCHI,CAC40
        //CIN,G93882192,Vodaphone
        //COM,001179594,Carrefour
        //ISN,CH0012138530,CS
        //SED,B1YW440,3i Group
        //VAL,24476758,UBS
        //WPK,A11Q05,Elumeo

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //Expose the extractionsContext from dssClient:
            ExtractionsContext extractionsContext = dssClient.extractionsContext;

            //-----------------------------------------------------------------
            //Check if the input file exists:
            //-----------------------------------------------------------------
            if (!File.Exists(instrumentIdentifiersInputFile))
            {
                DebugPrintAndWaitForEnter("ERROR accessing " + instrumentIdentifiersInputFile +
                    "\nCheck if file and directory exist.");
                return;  //Exit main program
            }

            //-----------------------------------------------------------------
            //Check if the output directory exists:
            //-----------------------------------------------------------------
            if (!Directory.Exists(outputDirectory))
            {
                DebugPrintAndWaitForEnter("ERROR accessing directory " + outputDirectory + "\nCheck if it exists.");
                return;  //Exit main program
            }

            //-----------------------------------------------------------------
            //Empty output files of data from previous run:
            //-----------------------------------------------------------------
            StreamWriter sw = new StreamWriter(dataOutputFile, false);
            sw.Close();
            Console.WriteLine("Cleared data output file " + dataOutputFile + "\n");

            sw = new StreamWriter(notesOutputFile, false);
            sw.Close();
            Console.WriteLine("Cleared notes output file " + notesOutputFile + "\n");

            sw = new StreamWriter(errorOutputFile, false);
            sw.Close();
            Console.WriteLine("Cleared error output file " + errorOutputFile + "\n");

            sw = new StreamWriter(gzipDataOutputFile, false);
            sw.Close();
            Console.WriteLine("Cleared compressed data output file " + gzipDataOutputFile + "\n");

            sw = new StreamWriter(gzipNotesOutputFile, false);
            sw.Close();
            Console.WriteLine("Cleared compressed data notes output file " + gzipNotesOutputFile + "\n");

            //-----------------------------------------------------------------
            //Again, we interface directly with the API.
            //This is why we do not create an instrument list and
            //report template on the DSS server.
            //-----------------------------------------------------------------

            //-----------------------------------------------------------------
            //Create an array outside of DSS, populate it with instruments,
            //read from an input file, with error handling and logging:
            //-----------------------------------------------------------------
            //Initialise the error output file:
            sw = new StreamWriter(errorOutputFile, true);
            sw.WriteLine("List of errors found in input file: " + instrumentIdentifiersInputFile + "\n");

            //Open the input file:
            StreamReader sr = new StreamReader(instrumentIdentifiersInputFile);

            //Instead of an array (of defined length), create a list for the instrument identifiers.
            //We do this because:
            //  we don't know how many instruments are in the file,
            //  we filter the file entries, to only keep the validated ones.
            //Once the list is filled, we convert it to an array, as that is what the API needs.
            //Create the empty instrument identifiers list:
            List<InstrumentIdentifier> instrumentIdentifiersList = new List<InstrumentIdentifier> ();

            //Populate the list, reading one line at a time from the file:
            int fileLineNumber = 0;
            string fileLine = string.Empty;
            string identifierTypeString = string.Empty;
            string identifierCodeString = string.Empty;
            IdentifierType identifierType;
            int i = 0;

            //Loop through all lines until we get to the end of the file:
            bool endOfFile = false;
            while (!endOfFile)
            {
                //Read one line of the file, test if end of file:
                fileLine = sr.ReadLine();
                endOfFile = (fileLine == null);
                if (!endOfFile)
                {
                    fileLineNumber++;
                    try
                    {
                        //Parse the file line to extract the comma separated instrument type and code:
                        string[] splitLine = fileLine.Split(new char[] { ',' });
                        identifierTypeString = splitLine[0];
                        identifierCodeString = splitLine[1];
                        //Add only validated instrument identifiers into our list.
                        if (identifierTypeString != string.Empty)
                        {
                            if (identifierCodeString != string.Empty)
                            {
                                //DSS can handle many types, here we only handle a subset:
                                switch (identifierTypeString)
                                {
                                    case "CHR": identifierType = IdentifierType.ChainRIC; break;
                                    case "CIN": identifierType = IdentifierType.Cin; break;
                                    case "COM": identifierType = IdentifierType.CommonCode; break;
                                    case "CSP": identifierType = IdentifierType.Cusip; break;
                                    case "ISN": identifierType = IdentifierType.Isin; break;
                                    case "RIC": identifierType = IdentifierType.Ric; break;
                                    case "SED": identifierType = IdentifierType.Sedol; break;
                                    case "VAL": identifierType = IdentifierType.Valoren; break;
                                    case "WPK": identifierType = IdentifierType.Wertpapier; break;
                                    default:
                                        identifierType = IdentifierType.NONE;
                                        DebugPrintAndWriteToFile("ERROR line " + fileLineNumber +
                                            ": unknown identifier type: " + identifierTypeString, sw);
                                        break;
                                }
                                if (identifierType != IdentifierType.NONE)
                                {
                                    instrumentIdentifiersList.Add(new InstrumentIdentifier
                                    {
                                        IdentifierType = identifierType,
                                        Identifier = identifierCodeString
                                    });
                                    Console.WriteLine("Line " + fileLineNumber + ": " + identifierTypeString +
                                        " " + identifierCodeString + " loaded into array [" + i + "]");
                                    i++;
                                }
                            }
                            //Error handling messages grouped here:
                            else
                            {
                                DebugPrintAndWriteToFile("ERROR line " + fileLineNumber +
                                    ": missing identifier code in line: " + fileLine, sw);
                            }
                        }
                        else
                        {
                            DebugPrintAndWriteToFile("ERROR line " + fileLineNumber +
                                ": missing identifier type in line: " + fileLine, sw);
                        }
                    }
                    catch
                    {
                        DebugPrintAndWriteToFile("ERROR line " + fileLineNumber +
                            ": bad line format: " + fileLine, sw);
                    }
                }
                else
                {
                    if (fileLineNumber == 0)
                        { DebugPrintAndWriteToFile("ERROR: empty file: " + instrumentIdentifiersInputFile, sw); }
                }
            }  //End of while loop

            sr.Close();
            sw.Close();

            int validIdentifiersCount = i;
            if (validIdentifiersCount == 0)
            {
                DebugPrintAndWaitForEnter("Exit program due to no identifiers in the list.");
                return;  //Exit main program
            }
            Console.WriteLine("\n" + validIdentifiersCount +
                " valid instruments were loaded into an array, outside of DSS,\n" +
                "for use in the extraction.");

            //Convert the instrument identifiers list to an array:
            InstrumentIdentifier[] instrumentIdentifiers = instrumentIdentifiersList.ToArray();

            //-----------------------------------------------------------------
            //Create an array of field names:
            //-----------------------------------------------------------------
            string[] requestedFieldNames = CreateRequestedFieldNames();


            //-----------------------------------------------------------------
            //Create and run an on demand extraction.
            //Variant 1: retrieve JSON formatted data, using ExtractWithNotes:
            //-----------------------------------------------------------------
            DebugPrintAndWaitForEnter("We also created an array of field names, outside of DSS." +
                "\n\nNext we will launch a direct on demand EOD pricing extraction," +
                "\nusing these 2 arrays.");

            Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
            ExtractionResult extractionResult =
                dssClient.CreateAndRunEodPricingExtraction(instrumentIdentifiers, requestedFieldNames);
            DssCollection<ExtractionRow> extractionDataRows = extractionResult.Contents;

            //-----------------------------------------------------------------
            //Data treatment starts here.
            //We display it on screen, and write it to a CSV file:
            //-----------------------------------------------------------------
            DisplayAndLogExtractedDataFieldNames(dataOutputFile, extractionDataRows);
            DisplayAndLogExtractedDataFieldValues(dataOutputFile, extractionDataRows);
            //-----------------------------------------------------------------
            //Extraction notes treatment starts here.
            //We analyze it, display it on screen, and write it to file:
            //-----------------------------------------------------------------
            DisplayAndLogAndAnalyzeExtractionNotes(notesOutputFile, errorOutputFile, extractionResult);


            //-----------------------------------------------------------------
            //Create and run an on demand extraction.
            //Variant 2: retrieve a Gzipped CSV, using ExtractRaw:
            //-----------------------------------------------------------------
            Console.WriteLine("We now run the same extraction, to retrieve Gzipped CSV data.");
            Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
            RawExtractionResult rawExtractionResult =
                dssClient.CreateAndRunEodPricingRawExtraction(instrumentIdentifiers, requestedFieldNames);

            //-----------------------------------------------------------------
            //Data treatment starts here.
            //We save the compressed data, as a gzip file:
            //-----------------------------------------------------------------
            //We must tell the server we accept compressed data, otherwise it will send uncompressed data:
            extractionsContext.DefaultRequestHeaders.Add("Accept-Encoding", "gzip");
            DssStreamResponse streamResponse = extractionsContext.GetReadStream(rawExtractionResult);
            using (FileStream fileStream = File.Create(gzipDataOutputFile))
                streamResponse.Stream.CopyTo(fileStream);
            extractionsContext.DefaultRequestHeaders.Remove("Accept-Encoding");
            Console.WriteLine("Saved the compressed data file to disk:\n" + gzipDataOutputFile);
            DebugPrintAndWaitForEnter("");
            //-----------------------------------------------------------------
            //Extraction notes treatment starts here.
            //We analyze it, display it on screen, and write it to file:
            //-----------------------------------------------------------------
            DisplayAndLogAndAnalyzeRawExtractionNotes(gzipNotesOutputFile, errorOutputFile, rawExtractionResult);
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        static void DebugPrintAndWriteToFile(string messageToPrintAndWriteToFile, StreamWriter sw)
        {
            Console.WriteLine(messageToPrintAndWriteToFile);
            sw.WriteLine(messageToPrintAndWriteToFile);
        }

        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = { "Instrument ID", "Security Description",
                                             "Universal Close Price Date", "Universal Close Price" };
            return requestedFieldNames;
        }

        static void DisplayAndLogExtractedDataFieldNames(
            string outputFile, DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            //Extract the names of the data fields, from the first row of returned data:
            StringBuilder returnedFieldNames = new StringBuilder(2048);
            ExtractionRow firstRow = extractionDataRows.First();
            foreach (string fieldName in firstRow.DynamicProperties.Keys)
            {
                returnedFieldNames.Append(fieldName);
                returnedFieldNames.Append(",");
            }
            DebugPrintAndWaitForEnter("Returned list of field names:\n" + returnedFieldNames.ToString());
            sw.WriteLine(returnedFieldNames.ToString());
            sw.Close();
        }

        static void DisplayAndLogExtractedDataFieldValues(
            string outputFile, DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            //Extract the actual data values, from all rows of returned data:
            Console.WriteLine("Returned field values:");
            Console.WriteLine("======================");
            string rowValuesInString = string.Empty;
            int validDataRows = 0;
            foreach (ExtractionRow row in extractionDataRows)
            {
                IEnumerable<object> rowValues = row.DynamicProperties.Select(dp => dp.Value);
                rowValuesInString = String.Join(", ", rowValues);
                if (rowValuesInString != string.Empty)
                {
                    Console.WriteLine(rowValuesInString);
                    sw.WriteLine(rowValuesInString);
                    validDataRows++;
                }
            }
            sw.Close();

            //The number of data rows could be less than the number of instruments,
            //if data is not available for some instruments:
            DebugPrintAndWaitForEnter("\nExtraction completed." +
                "\nNumber of data rows: " + extractionDataRows.Count() +
                "\nNumber of valid (non empty) data rows: " + validDataRows +
                "\n\nData was also written to file.");
        }

        static void DisplayAndLogExtractionNotes(string outputFile, ExtractionResult extractionResult)
        {
            //Error handling:
            if (!extractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");
            foreach (string note in extractionResult.Notes)
            {
                Console.WriteLine(note);
                sw.WriteLine(note);
            }
            sw.Close();
            DebugPrintAndWaitForEnter("\nThese notes were also written to file");
        }

        static void DisplayAndLogAndAnalyzeExtractionNotes(string outputFile, string errorOutputFile, ExtractionResult extractionResult)
        {
            //Error handling:
            if (!extractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string inactiveMsgs = "";
            string invalidMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in extractionResult.Notes)
            {
                Console.WriteLine(notes);
                sw.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("inactive")) { inactiveMsgs = inactiveMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("invalid")) { invalidMsgs = invalidMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            sw.Close();
            Console.WriteLine("===============================================================================");

            StreamWriter swErr = new StreamWriter(errorOutputFile, true);
            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                swErr.WriteLine("\nERROR messages in notes file:\n" + errorMsgs);
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                swErr.WriteLine("WARNING messages in notes file:\n" + warningMsgs);
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (inactiveMsgs != "")
            {
                swErr.WriteLine("Inactive instruments messages in notes file:\n" + inactiveMsgs);
                Console.WriteLine("Inactive instruments messages:\n" + inactiveMsgs);
            }
            if (invalidMsgs != "")
            {
                swErr.WriteLine("Invalid instruments messages in notes file:\n" + invalidMsgs);
                Console.WriteLine("Invalid instruments messages:\n" + invalidMsgs);
            }
            if (permissionMsgs != "")
            {
                swErr.WriteLine("PERMISSION ISSUES messages in notes file:\n" + permissionMsgs);
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            swErr.Close();
            DebugPrintAndWaitForEnter("===============================================================================");
        }

        static void DisplayAndLogAndAnalyzeRawExtractionNotes(string outputFile, string errorOutputFile, RawExtractionResult rawExtractionResult)
        {
            //Error handling:
            if (!rawExtractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in rawExtractionResult.Notes)
            {
                Console.WriteLine(notes);
                sw.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            sw.Close();
            Console.WriteLine("===============================================================================");

            StreamWriter swErr = new StreamWriter(errorOutputFile, true);
            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                swErr.WriteLine("\nERROR messages in notes file:\n" + errorMsgs);
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                swErr.WriteLine("WARNING messages in notes file:\n" + warningMsgs);
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (permissionMsgs != "")
            {
                swErr.WriteLine("PERMISSION ISSUES messages in notes file:\n" + permissionMsgs);
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            swErr.Close();
            DebugPrintAndWaitForEnter("===============================================================================");
        }
    }
}