//============================================================================
//DSS REST API Learning Tutorial 7: main program code
//Goal:  continue productizing our code:
//       instrument lists: create a large instrument list array outside of DSS,
//                         populated from an input file, with error handling;
//       field lists: create a field list array, outside of DSS;
//       We show 2 variants:
//       1) create and run an EoD on demand extraction, for JSON formatted data;
//       wait for it to complete;
//       display received data on screen and write it to a CSV file;
//       display received extraction notes on screen and write them to file;
//       analyze the extraction notes;
//       include some error checking.
//       2) repeat for GZipped CSV formatted data (use case: large data sets);
//       in this scenario we do not display the data on screen,
//       we just save it to a Gzipped CSV file;
//       Note) Cleanup is not required in this case.
//Limit: this code is the same as Tutorial 6, except that it handles any
//       instrument list size, even if it is greater than the maximum size
//       of DSS instrument lists (75000 instruments).
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ExtractionRequests;
using DataScope.Select.Core.RestApi;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        private static int maxDssIdentifiersListSize = 75000;

        private static string outputDirectory = "C:\\DSS REST API\\";
        private static string dataOutputFile = outputDirectory + "DSS_API_tutorial_7l_output.txt";
        private static string notesOutputFile = outputDirectory + "DSS_API_tutorial_7l_notes.txt";
        private static string errorOutputFile = outputDirectory + "DSS_API_tutorial_7l_errors.txt";
        private static string gzipDataOutputFile = outputDirectory + "DSS_API_tutorial_7l_gzipOutput.csv.gz";
        private static string gzipNotesOutputFile = outputDirectory + "DSS_API_tutorial_7l_gzipNotes.txt";

        private static string inputDirectory = outputDirectory;
        private static string instrumentIdentifiersInputFile = inputDirectory + "DSS_API_10_input_file.csv";
        //File format: 1 instrument per line.
        //Line format: <identifier type>, <code>, <optional comment>
        //Example lines, first two are for a Cusip (CSP) and a RIC:
        //CSP,31352JRV1
        //RIC,ALVG.DE
        //CHR,0#.FCHI,CAC40
        //CIN,G93882192,Vodaphone
        //COM,001179594,Carrefour
        //ISN,CH0012138530,CS
        //SED,B1YW440,3i Group
        //VAL,24476758,UBS
        //WPK,A11Q05,Elumeo

        //List of possible input file errors, used in error handling and reporting:
        private enum Errors
            { NoError, EmptyFile, BadLineFormat, EmptyType, EmptyCode, UnknownIdentifier };

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();
            
            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //-----------------------------------------------------------------
            //Check if the input file exists:
            //-----------------------------------------------------------------
            if (!FileExists(instrumentIdentifiersInputFile)) { return; }  //Exit main program

            //-----------------------------------------------------------------
            //Check if the output directory exists:
            //-----------------------------------------------------------------
            if (!DirectoryExists(outputDirectory)) { return; }  //Exit main program

            //-----------------------------------------------------------------
            //Empty output files of data from previous run:
            //-----------------------------------------------------------------
            ClearFileContents(dataOutputFile);
            Console.WriteLine("Cleared data output file " + dataOutputFile + "\n");
            ClearFileContents(notesOutputFile);
            Console.WriteLine("Cleared notes output file " + notesOutputFile + "\n");
            ClearFileContents(errorOutputFile);
            Console.WriteLine("Cleared error output file " + errorOutputFile + "\n");
            ClearFileContents(gzipDataOutputFile);
            Console.WriteLine("Cleared compressed data output file " + gzipDataOutputFile + "\n");
            ClearFileContents(gzipNotesOutputFile);
            DebugPrintAndWaitForEnter("Cleared compressed data notes output file " + gzipNotesOutputFile + "\n");

            //-----------------------------------------------------------------
            //Again, we interface directly with the API.
            //This is why we do not create an instrument list and
            //report template on the DSS server.
            //-----------------------------------------------------------------

            //-----------------------------------------------------------------
            //Create a list outside of DSS, populate it with instruments,
            //read from an input file, with error handling and logging:
            //-----------------------------------------------------------------
            //Instead of an array (of defined length), we create a list for the instrument identifiers.
            //We do this because:
            //  we don't know how many instruments are in the file,
            //  we filter the file entries, to only keep the validated ones.
            List<InstrumentIdentifier> instrumentIdentifiersList =
                PopulateInstrumentIdentifiersListFromFile(instrumentIdentifiersInputFile, errorOutputFile);
            int validIdentifiersCount = instrumentIdentifiersList.Count();
            if (validIdentifiersCount == 0)
            {
                DebugPrintAndWaitForEnter("Exit program due to no identifiers in the list.");
                return;  //Exit main program
            }
            Console.WriteLine("\n" + validIdentifiersCount +
                " valid instruments were loaded into a list, outside of DSS,\n" +
                "for use in the extraction.");

            //-----------------------------------------------------------------
            //Create an array of field names:
            //-----------------------------------------------------------------
            string[] requestedFieldNames = CreateRequestedFieldNames();
            Console.WriteLine("We also created an array of field names, outside of DSS.\n");

            //-----------------------------------------------------------------
            //Create and run one (or more) on demand extraction(s),
            //each followed by data treatment, inside a big loop.
            //DSS has a limit on the number of instruments in a list.
            //That limit is 75000, set in variable maxDssIdentifiersListSize.
            //If the file contains more, we must split the list into:
            //  1 or more chunks of 75000
            //  an eventual remainder
            //We run an extraction and treat the data for each sub-list, inside a big loop.
            //Inside each iteration of the loop:
            //  Create an instrument identifier list (a subset of the main one)
            //  Create an on demand EOD extraction for that instrument identifier list
            //  Display the data on screen and write it to file
            //-----------------------------------------------------------------
            //The code determines how many loops we need to run, taking all cases into account:
            //  A list divisible by 75000, without remainder.
            //  A list divisible by 75000, with a remainder.
            //Determine the number of loops of 75000 to run:
            int loopsToRun = validIdentifiersCount / maxDssIdentifiersListSize;
            //Determine the number of instruments in the remainder loop:
            int sizeOfInstrumentIdentifiersRemainder = validIdentifiersCount % maxDssIdentifiersListSize;
            //Add a loop for the remainder:
            if (sizeOfInstrumentIdentifiersRemainder > 0) { loopsToRun++; }

            Console.WriteLine("Next we will launch " + loopsToRun +
                " on demand EOD pricing extraction(s),\n" +
                "each followed by data treatment.\n" +
                "The " + loopsToRun + " iteration(s) take into account the " +
                maxDssIdentifiersListSize + " instruments limit.");
            if (sizeOfInstrumentIdentifiersRemainder > 0)
            {
                DebugPrintAndWaitForEnter("The last iteration has " + sizeOfInstrumentIdentifiersRemainder + " instruments.\n");
            }
            else
            {
                DebugPrintAndWaitForEnter("The last iteration has " + maxDssIdentifiersListSize + " instruments.\n");
            }

            DssCollection<ExtractionRow> extractionDataRows;
            int bigLoopIteration = 1;
            int firstListElementIndex = 0;
            int numberOfListElements = 0;

            //This is the big loop that will run as many times as required to handle all instruments:
            while (bigLoopIteration <= loopsToRun)
            {
                //Fill the instrument identifiers subset array with a set of instruments,
                //either 75000, or less if it is the remainder:
                firstListElementIndex = (bigLoopIteration - 1) * maxDssIdentifiersListSize;
                
                if (bigLoopIteration < loopsToRun || sizeOfInstrumentIdentifiersRemainder == 0)
                    { numberOfListElements = maxDssIdentifiersListSize; }
                else
                    { numberOfListElements = sizeOfInstrumentIdentifiersRemainder; }

                List<InstrumentIdentifier> instrumentIdentifiersSubList =
                    instrumentIdentifiersList.GetRange(firstListElementIndex, numberOfListElements);

                //Convert the sublist to an array, as that is what the API needs.
                InstrumentIdentifier[] instrumentIdentifiersSubArray = instrumentIdentifiersSubList.ToArray();
                Console.WriteLine("Loaded identifiers into a sub array, iteration number: " + bigLoopIteration +
                    "\nFirst element index: " + firstListElementIndex + " - Number of elements: " + numberOfListElements);


                //-----------------------------------------------------------------
                //Create and run an on demand extraction, using our instrument identifiers subset array:
                //Variant 1: retrieve JSON formatted data, using ExtractWithNotes:
                //-----------------------------------------------------------------
                Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
                ExtractionResult extractionResult =
                    dssClient.CreateAndRunEodPricingExtraction(instrumentIdentifiersSubArray, requestedFieldNames);
                extractionDataRows = extractionResult.Contents;

                //-----------------------------------------------------------------
                //Data treatment starts here.
                //We display it on screen, write it to file, and check the amount of data:
                //-----------------------------------------------------------------
                if (bigLoopIteration == 1)
                    { DisplayAndLogExtractedDataFieldNames(dataOutputFile, extractionDataRows); }
                DisplayAndLogExtractedDataFieldValues(dataOutputFile, extractionDataRows);
                DisplayAndLogAndAnalyzeExtractionNotes(notesOutputFile, errorOutputFile, extractionResult);


                //-----------------------------------------------------------------
                //Create and run an on demand extraction.
                //Variant 2: retrieve a Gzipped CSV, using ExtractRaw:
                //-----------------------------------------------------------------
                Console.WriteLine("We now run the same extraction, to retrieve Gzipped CSV data.");
                Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
                RawExtractionResult rawExtractionResult =
                    dssClient.CreateAndRunEodPricingRawExtraction(instrumentIdentifiersSubArray, requestedFieldNames);

                //-----------------------------------------------------------------
                //Data treatment starts here.
                //We save the compressed data, as a gzip file:
                //-----------------------------------------------------------------
                dssClient.SaveCompressedData(rawExtractionResult, gzipDataOutputFile);
                DebugPrintAndWaitForEnter("");
                //-----------------------------------------------------------------
                //Extraction notes treatment starts here.
                //We analyze it, display it on screen, and write it to file:
                //-----------------------------------------------------------------
                DisplayAndLogAndAnalyzeRawExtractionNotes(gzipNotesOutputFile, errorOutputFile, rawExtractionResult);


                //Increment the big while loop to process the next set of identifiers:
                bigLoopIteration++;
            }

            DebugPrintAndWaitForEnter("All extractions completed.");
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        static void DebugPrintAndWriteToFile(string messageToPrintAndWriteToFile, StreamWriter sw)
        {
            Console.WriteLine(messageToPrintAndWriteToFile);
            sw.WriteLine(messageToPrintAndWriteToFile);
        }

        static void DebugPrintAndWriteToFileErrorMessage(
            Errors errorCode, int fileLineNumber, string fileLine, string identifierTypeString, StreamWriter sw)
        {
            string errorMessage = string.Empty;
            switch (errorCode)
            {
                case Errors.NoError:
                    break;
                case Errors.EmptyFile:
                    errorMessage = "ERROR: empty file: " + instrumentIdentifiersInputFile;
                    break;
                case Errors.BadLineFormat:
                    errorMessage = "ERROR line " + fileLineNumber + ": bad line format: " + fileLine;
                    break;
                case Errors.EmptyType:
                    errorMessage = "ERROR line " + fileLineNumber + ": missing identifier type in line: " + fileLine;
                    break;
                case Errors.EmptyCode:
                    errorMessage = "ERROR line " + fileLineNumber + ": missing identifier code in line: " + fileLine;
                    break;
                case Errors.UnknownIdentifier:
                    errorMessage = "ERROR line " + fileLineNumber + ": unknown identifier type: " + identifierTypeString;
                    break;
                default:
                    errorMessage = "ERROR in code: unknown error: " + errorCode;
                    break;
            }
            DebugPrintAndWriteToFile(errorMessage, sw);
        }

        static bool FileExists(string fileName)
        {
            if (!File.Exists(fileName))
            {
                DebugPrintAndWaitForEnter("ERROR accessing " + fileName +
                    "\nCheck if file and directory exist.");
                return false;
            }
            else { return true; }
        }

        static bool DirectoryExists(string dirName)
        {
            if (!Directory.Exists(dirName))
            {
                DebugPrintAndWaitForEnter("ERROR accessing directory " + dirName + "\nCheck if it exists.");
                return false;
            }
            else { return true; }
        }

        //Clear contents of a file to ensure it does not contain data from a previous run:
        static void ClearFileContents(string fileName)
        {
            StreamWriter sw = new StreamWriter(fileName, false);
            sw.Close();
        }

        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = { "Instrument ID", "Security Description",
                                             "Universal Close Price Date", "Universal Close Price" };
            return requestedFieldNames;
        }

        //Create a list outside of DSS, populate it with instruments,
        //read from an input file, with error handling and logging:        
        static List<InstrumentIdentifier> PopulateInstrumentIdentifiersListFromFile(
            string instrumentIdentifiersInputFile, string errorOutputFile)
        {
            //Open the input file:
            StreamReader sr = new StreamReader(instrumentIdentifiersInputFile);

            //Initialise the error output file:
            StreamWriter sw = new StreamWriter(errorOutputFile, true);
            sw.WriteLine("List of errors found in input file: " + instrumentIdentifiersInputFile + "\n");

            //Instead of an array (of defined length), create a list for the instrument identifiers.
            //We do this because:
            //  we don't know how many instruments are in the file,
            //  we filter the file entries, to only keep the validated ones.
            //Once the list is filled, we convert it to an array, as that is what the API needs.
            //Create the empty instrument identifiers list:
            List<InstrumentIdentifier> instrumentIdentifiersList = new List<InstrumentIdentifier> ();

            //Populate the list, reading one line at a time from the file:
            int fileLineNumber = 0;
            string fileLine = string.Empty;
            bool commaExistsInFileLine = false;
            string identifierTypeString = string.Empty;
            string identifierCodeString = string.Empty;
            IdentifierType identifierType;
            int i = 0;

            //Loop through all lines until we get to the end of the file:
            bool endOfFile = false;
            while (!endOfFile)
            {
                Errors errorCode = Errors.NoError;

                //Read one line of the file, test if end of file:
                fileLine = sr.ReadLine();
                endOfFile = (fileLine == null);
                if (endOfFile && fileLineNumber == 0) { errorCode = Errors.EmptyFile; };
                fileLineNumber++;

                //Parse the file line to extract the comma separated instrument type and code:
                if (errorCode == Errors.NoError && !endOfFile)
                {
                    commaExistsInFileLine = (fileLine.IndexOf(",") >= 0);
                    if (commaExistsInFileLine)
                    {
                        string[] splitLine = fileLine.Split(new char[] { ',' });
                        identifierTypeString = splitLine[0];
                        identifierCodeString = splitLine[1];
                    }
                    else
                    {
                        errorCode = Errors.BadLineFormat;  //Missing comma
                        identifierTypeString = string.Empty;
                        identifierCodeString = string.Empty;
                    }
                }
                if (identifierTypeString == string.Empty && errorCode == Errors.NoError)
                    { errorCode = Errors.EmptyType; }
                if (identifierCodeString == string.Empty && errorCode == Errors.NoError)
                    { errorCode = Errors.EmptyCode; }

                identifierType = IdentifierType.NONE;
                if (errorCode == Errors.NoError && !endOfFile)
                {
                    //DSS can handle many types, here we only handle a subset:
                    switch (identifierTypeString)
                    {
                        case "CHR": identifierType = IdentifierType.ChainRIC; break;
                        case "CIN": identifierType = IdentifierType.Cin; break;
                        case "COM": identifierType = IdentifierType.CommonCode; break;
                        case "CSP": identifierType = IdentifierType.Cusip; break;
                        case "ISN": identifierType = IdentifierType.Isin; break;
                        case "RIC": identifierType = IdentifierType.Ric; break;
                        case "SED": identifierType = IdentifierType.Sedol; break;
                        case "VAL": identifierType = IdentifierType.Valoren; break;
                        case "WPK": identifierType = IdentifierType.Wertpapier; break;
                        default: errorCode = Errors.UnknownIdentifier; break;
                    }
                }

                if (errorCode == Errors.NoError && !endOfFile)
                {
                    //Add validated instrument identifier into our list:
                    instrumentIdentifiersList.Add(new InstrumentIdentifier
                    {
                        IdentifierType = identifierType,
                        Identifier = identifierCodeString
                    });
                    Console.WriteLine("Line " + fileLineNumber + ": " +
                        identifierTypeString + " " + identifierCodeString + " loaded into array [" + i + "]");
                    i++;
                }

                if (errorCode != Errors.NoError)
                {
                    DebugPrintAndWriteToFileErrorMessage(
                                 errorCode, fileLineNumber, fileLine, identifierTypeString, sw);
                }
            }  //End of while loop

            sr.Close();
            sw.Close();

            return instrumentIdentifiersList;
        }

        static void DisplayAndLogExtractedDataFieldNames(
            string outputFile, DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            //Extract the names of the data fields, from the first row of returned data:
            StringBuilder returnedFieldNames = new StringBuilder(2048);
            ExtractionRow firstRow = extractionDataRows.First();
            foreach (string fieldName in firstRow.DynamicProperties.Keys)
            {
                returnedFieldNames.Append(fieldName);
                returnedFieldNames.Append(",");
            }
            DebugPrintAndWaitForEnter("Returned list of field names:\n" + returnedFieldNames.ToString());
            sw.WriteLine(returnedFieldNames.ToString());
            sw.Close();
        }

        static void DisplayAndLogExtractedDataFieldValues(
            string outputFile, DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            //Extract the actual data values, from all rows of returned data:
            Console.WriteLine("Returned field values:");
            Console.WriteLine("======================");
            string rowValuesInString = string.Empty;
            int validDataRows = 0;
            foreach (ExtractionRow row in extractionDataRows)
            {
                IEnumerable<object> rowValues = row.DynamicProperties.Select(dp => dp.Value);
                rowValuesInString = String.Join(", ", rowValues);
                if (rowValuesInString != string.Empty)
                {
                    Console.WriteLine(rowValuesInString);
                    sw.WriteLine(rowValuesInString);
                    validDataRows++;
                }
            }

            //The number of data rows could be less than the number of instruments,
            //if data is not available for some instruments:
            DebugPrintAndWaitForEnter("\nExtraction completed." +
                "\nNumber of data rows: " + extractionDataRows.Count() +
                "\nNumber of valid (non empty) data rows: " + validDataRows +
                "\n\nOutput was also written to file.");
            sw.Close();
        }

        static void DisplayAndLogExtractionNotes(string outputFile, ExtractionResult extractionResult)
        {
            //Error handling:
            if (!extractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");
            foreach (string note in extractionResult.Notes)
            {
                Console.WriteLine(note);
                sw.WriteLine(note);
            }
            sw.WriteLine("===============================================================================");
            sw.Close();
            DebugPrintAndWaitForEnter("\nThese notes were also written to file");
        }

        static void DisplayAndLogAndAnalyzeExtractionNotes(string outputFile, string errorOutputFile, ExtractionResult rawExtractionResult)
        {
            //Error handling:
            if (!rawExtractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string inactiveMsgs = "";
            string invalidMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in rawExtractionResult.Notes)
            {
                Console.WriteLine(notes);
                sw.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("inactive")) { inactiveMsgs = inactiveMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("invalid")) { invalidMsgs = invalidMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            sw.Close();
            Console.WriteLine("===============================================================================");

            StreamWriter swErr = new StreamWriter(errorOutputFile, true);
            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                swErr.WriteLine("\nERROR messages in notes file:\n" + errorMsgs);
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                swErr.WriteLine("WARNING messages in notes file:\n" + warningMsgs);
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (inactiveMsgs != "")
            {
                swErr.WriteLine("Inactive instruments messages in notes file:\n" + inactiveMsgs);
                Console.WriteLine("Inactive instruments messages:\n" + inactiveMsgs);
            }
            if (invalidMsgs != "")
            {
                swErr.WriteLine("Invalid instruments messages in notes file:\n" + invalidMsgs);
                Console.WriteLine("Invalid instruments messages:\n" + invalidMsgs);
            }
            if (permissionMsgs != "")
            {
                swErr.WriteLine("PERMISSION ISSUES messages in notes file:\n" + permissionMsgs);
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            swErr.Close();
            DebugPrintAndWaitForEnter("===============================================================================");
        }

        static void DisplayAndLogAndAnalyzeRawExtractionNotes(string outputFile, string errorOutputFile, RawExtractionResult rawExtractionResult)
        {
            //Error handling:
            if (!rawExtractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in rawExtractionResult.Notes)
            {
                Console.WriteLine(notes);
                sw.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            sw.Close();
            Console.WriteLine("===============================================================================");

            StreamWriter swErr = new StreamWriter(errorOutputFile, true);
            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                swErr.WriteLine("\nERROR messages in notes file:\n" + errorMsgs);
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                swErr.WriteLine("WARNING messages in notes file:\n" + warningMsgs);
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (permissionMsgs != "")
            {
                swErr.WriteLine("PERMISSION ISSUES messages in notes file:\n" + permissionMsgs);
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            swErr.Close();
            DebugPrintAndWaitForEnter("===============================================================================");
        }
    }
}
