﻿//============================================================================
//DSS REST API Learning Tutorial 8: main program code
//Goal: run an On Demand extraction for terms and conditions data.
//      In this example we cross reference instrument codes, i.e. retrieve
//      several instrument codes (RIC; Cusip, ISIN, Sedol, etc.), for a list
//      of instruments (with different instrument codes).
//      For each input instrument code, only 1 RIC is returned (the primary RIC).
//      We also output and analyze the extraction notes.
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ExtractionRequests;
using DataScope.Select.Core.RestApi;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //-----------------------------------------------------------------
            //Again, we interface directly with the API, so we do not create
            //an instrument list and report template on the DSS server.
            //-----------------------------------------------------------------

            //-----------------------------------------------------------------
            //Create an array outside of DSS, populate it with instruments:
            //-----------------------------------------------------------------
            //In this example we do something very simple (like in tutorial 5),
            //to concentrate on the extraction itself.
            //For a more productized example with details on handling large
            //instrument lists, read from file, see tutorial 7.
            InstrumentIdentifier[] instrumentIdentifiers = PopulateInstrumentIdentifiers();

            //-----------------------------------------------------------------
            //Create an array of field names:
            //-----------------------------------------------------------------
            //We do not create a report template, we only create an array of field names.
            //We will use the array in the extractions.
            string[] requestedFieldNames = CreateRequestedFieldNames();

            //-----------------------------------------------------------------
            //Create and run an on demand extraction:
            //-----------------------------------------------------------------
            //We choose to do a T&C (Terms and Conditions) extraction:
            DebugPrintAndWaitForEnter("We created an array with " + instrumentIdentifiers.Length + " instrument identifiers, " +
                "and an array of field names, outside of DSS." +
                "\n\nNext we will launch a direct on demand T&C extraction," +
                "\nusing these 2 arrays.");

            Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
            ExtractionResult extractionResult =
                dssClient.CreateAndRunTandCExtraction(instrumentIdentifiers, requestedFieldNames);
            DssCollection<ExtractionRow> extractionDataRows = extractionResult.Contents;

            //-----------------------------------------------------------------
            //Data treatment starts here. We just display it on screen:
            //-----------------------------------------------------------------
            DisplayExtractedDataFieldNames(extractionDataRows);
            DisplayExtractedDataFieldValues(extractionDataRows);
            DisplayAndAnalyzeExtractionNotes(extractionResult);
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        static string[] CreateRequestedFieldNames()
        {
            string[] requestedFieldNames = { "RIC", "CUSIP", "ISIN", "SEDOL", "Issuer OrgID",
                                             "Currency Code" };
            return requestedFieldNames;
        }

        //Create an array for the instrument identifiers, populate it manually:
        static InstrumentIdentifier[] PopulateInstrumentIdentifiers()
        {
            //Create an array for the instrument identifiers, of size 5:
            InstrumentIdentifier[] instrumentIdentifiers = new InstrumentIdentifier[5];

            //In this example we populate it manually:
            instrumentIdentifiers[0] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Cusip,
                Identifier = "00209tab1"
            };
            instrumentIdentifiers[1] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Ric,
                Identifier = "IBM.N"
            };
            instrumentIdentifiers[2] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Isin,
                Identifier = "US4592001014"
            };
            instrumentIdentifiers[3] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Sedol,
                Identifier = "B1YW440"
            };
            //Inactive instrument:
            instrumentIdentifiers[4] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Cusip,
                Identifier = "438516AC0"
            };
            return instrumentIdentifiers;
        }

        //Extract the names of the data fields, from the first row of returned data:
        static void DisplayExtractedDataFieldNames(DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            //Extract the names of the data fields, from the first row of returned data:
            StringBuilder returnedFieldNames = new StringBuilder();
            ExtractionRow firstRow = extractionDataRows.First();
            foreach (string fieldName in firstRow.DynamicProperties.Keys)
            {
                returnedFieldNames.Append(fieldName);
                returnedFieldNames.Append(",");
            }
            DebugPrintAndWaitForEnter("Returned list of field names:\n" + returnedFieldNames.ToString());
        }

        //Extract the actual data values, from all rows of returned data:
        static void DisplayExtractedDataFieldValues(DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            //Extract the actual data values, from all rows of returned data:
            Console.WriteLine("Returned field values:");
            Console.WriteLine("======================");
            string rowValuesInString = string.Empty;
            int validDataRows = 0;
            foreach (ExtractionRow row in extractionDataRows)
            {
                IEnumerable<object> rowValues = row.DynamicProperties.Select(dp => dp.Value);
                rowValuesInString = String.Join(", ", rowValues);
                if (rowValuesInString != string.Empty)
                {
                    Console.WriteLine(rowValuesInString);
                    validDataRows++;
                }
            }

            //The number of data rows could be less than the number of instruments,
            //if data is not available for some instruments:
            DebugPrintAndWaitForEnter("\nExtraction completed." +
                "\nNumber of data rows: " + extractionDataRows.Count() +
                "\nNumber of valid (non empty) data rows: " + validDataRows);
        }

        //Display the extraction notes:
        static void DisplayExtractionNotes(ExtractionResult extractionResult)
        {
            //Error handling:
            if (!extractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");
            foreach (string note in extractionResult.Notes)
                Console.WriteLine(note);
            DebugPrintAndWaitForEnter("");
        }

        static void DisplayAndAnalyzeExtractionNotes(ExtractionResult rawExtractionResult)
        {
            //Error handling:
            if (!rawExtractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string inactiveMsgs = "";
            string invalidMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in rawExtractionResult.Notes)
            {
                Console.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("inactive")) { inactiveMsgs = inactiveMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("invalid")) { invalidMsgs = invalidMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            Console.WriteLine("===============================================================================");

            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (inactiveMsgs != "")
            {
                Console.WriteLine("Inactive instruments messages:\n" + inactiveMsgs);
            }
            if (invalidMsgs != "")
            {
                Console.WriteLine("Invalid instruments messages:\n" + invalidMsgs);
            }
            if (permissionMsgs != "")
            {
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            DebugPrintAndWaitForEnter("===============================================================================");
        }
    }
}
