﻿//============================================================================
//DSS REST API Learning Tutorial 9: main program code
//Goal: run an On Demand extraction for composite data.
//      Composite extractions combine T&C and pricing data.
//      In this example we retrieve a few composite fields related to
//      annualized dividends and the balance sheet.
//      We also cross reference instrument codes, i.e. retrieve several
//      instrument codes (RIC; Cusip, ISIN, Sedol, etc.), for a list of
//      instruments (with different instrument codes).
//      For each input instrument code, only 1 RIC is returned (the primary RIC).
//      We also output and analyze the extraction notes.
//      We show 2 variants:
//      1) create and run an EoD on demand extraction, for JSON formatted data;
//      2) repeat for GZipped CSV formatted data (use case: large data sets);
//      in this scenario we do not display the data on screen,
//      we just save it to a Gzipped CSV file;
//============================================================================
//To facilitate the learning process, the member declarations are at the start,
//followed by the main and finally the methods.
//============================================================================
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using DataScope.Select.Api.Content;
using DataScope.Select.Api.Extractions.ExtractionRequests;
using DataScope.Select.Core.RestApi;
using DssRestfulApiTutorials;

namespace DssRestfulApiTutorials
{
    class Program
    {
        //=====================================================================
        //Member declarations
        //=====================================================================
        private static string dssUserName = "YourUserId";
        private static string dssUserPassword = "YourPassword";

        private static string outputDirectory = "C:\\DSS REST API\\";
        private static string dataOutputFile = outputDirectory + "DSS_API_tutorial_9l_output.csv";
        private static string notesOutputFile = outputDirectory + "DSS_API_tutorial_9l_notes.txt";
        private static string errorOutputFile = outputDirectory + "DSS_API_tutorial_9l_errors.txt";
        private static string gzipDataOutputFile = outputDirectory + "DSS_API_tutorial_9l_gzipOutput.csv.gz";
        private static string gzipNotesOutputFile = outputDirectory + "DSS_API_tutorial_9l_gzipNotes.txt";

        private static string inputDirectory = outputDirectory;

        //=====================================================================
        //Main program entry
        //=====================================================================
        static void Main()
        {
            //-----------------------------------------------------------------
            //Connect and authenticate to the DSS server:
            //-----------------------------------------------------------------
            DssClient dssClient = new DssClient();

            dssClient.ConnectToServer(dssUserName, dssUserPassword);
            DebugPrintAndWaitForEnter("Returned session token: " + dssClient.SessionToken);

            //-----------------------------------------------------------------
            //Check if the output directory exists:
            //-----------------------------------------------------------------
            if (!DirectoryExists(outputDirectory)) { return; }  //Exit main program

            //-----------------------------------------------------------------
            //Empty output files of data from previous run:
            //-----------------------------------------------------------------
            ClearFileContents(dataOutputFile);
            Console.WriteLine("Cleared data output file " + dataOutputFile + "\n");
            ClearFileContents(notesOutputFile);
            Console.WriteLine("Cleared notes output file " + notesOutputFile + "\n");
            ClearFileContents(errorOutputFile);
            Console.WriteLine("Cleared error output file " + errorOutputFile + "\n");
            ClearFileContents(gzipDataOutputFile);
            Console.WriteLine("Cleared compressed data output file " + gzipDataOutputFile + "\n");
            ClearFileContents(gzipNotesOutputFile);
            DebugPrintAndWaitForEnter("Cleared compressed data notes output file " + gzipNotesOutputFile + "\n");

            //-----------------------------------------------------------------
            //Again, we interface directly with the API, so we do not create
            //an instrument list and report template on the DSS server.
            //-----------------------------------------------------------------

            //-----------------------------------------------------------------
            //Create an array outside of DSS, populate it with instruments:
            //-----------------------------------------------------------------
            //In this example we do something very simple (like in tutorial 5),
            //to concentrate on the extraction itself.
            //For a more productized example with details on handling large
            //instrument lists, read from file, see tutorial 7.
            InstrumentIdentifier[] instrumentIdentifiers = PopulateInstrumentIdentifiers();

            //-----------------------------------------------------------------
            //Create an array of field names:
            //-----------------------------------------------------------------
            //We do not create a report template, we only create an array of field names.
            //We will use the array in the extractions.
            string[] requestedFieldNames = CreateRequestedFieldNames();


            //-----------------------------------------------------------------
            //Create and run an on demand extraction:
            //Variant 1: retrieve JSON formatted data, using ExtractWithNotes:
            //-----------------------------------------------------------------
            DebugPrintAndWaitForEnter("We created an array with " + instrumentIdentifiers.Length + " instrument identifiers, " +
                "and an array of field names, outside of DSS." +
                "\n\nNext we will launch a direct on demand composite (T&C and pricing) extraction," +
                "\nusing these 2 arrays.");

            Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
            ExtractionResult extractionResult =
                dssClient.CreateAndRunCompositeExtraction(instrumentIdentifiers, requestedFieldNames);
            DssCollection<ExtractionRow> extractionDataRows = extractionResult.Contents;

            //-----------------------------------------------------------------
            //Data treatment starts here.
            //We display it on screen, and write it to a CSV file:
            //-----------------------------------------------------------------
            DisplayAndLogExtractedDataFieldNames(dataOutputFile, extractionDataRows);
            DisplayAndLogExtractedDataFieldValues(dataOutputFile, extractionDataRows);
            //-----------------------------------------------------------------
            //Extraction notes treatment starts here.
            //We analyze it, display it on screen, and write it to file:
            //-----------------------------------------------------------------
            DisplayAndLogAndAnalyzeExtractionNotes(notesOutputFile, errorOutputFile, extractionResult);


            //-----------------------------------------------------------------
            //Create and run an on demand extraction.
            //Variant 2: retrieve a Gzipped CSV, using ExtractRaw:
            //-----------------------------------------------------------------
            Console.WriteLine("We now run the same extraction, to retrieve Gzipped CSV data.");
            Console.WriteLine("Please be patient and wait for the extraction to complete ...\n");
            RawExtractionResult rawExtractionResult =
                dssClient.CreateAndRunCompositeRawExtraction(instrumentIdentifiers, requestedFieldNames);

            //-----------------------------------------------------------------
            //Data treatment starts here.
            //We save the compressed data, as a gzip file:
            //-----------------------------------------------------------------
            dssClient.SaveCompressedData(rawExtractionResult, gzipDataOutputFile);
            DebugPrintAndWaitForEnter("");
            //-----------------------------------------------------------------
            //Extraction notes treatment starts here.
            //We analyze it, display it on screen, and write it to file:
            //-----------------------------------------------------------------
            DisplayAndLogAndAnalyzeRawExtractionNotes(gzipNotesOutputFile, errorOutputFile, rawExtractionResult);
        }

        //=====================================================================
        //Helper methods
        //=====================================================================
        static void DebugPrintAndWaitForEnter(string messageToPrint)
        {
            Console.WriteLine(messageToPrint);
            Console.WriteLine("Press Enter to continue");
            Console.ReadLine();
        }

        static bool DirectoryExists(string dirName)
        {
            if (!Directory.Exists(dirName))
            {
                DebugPrintAndWaitForEnter("ERROR accessing directory " + dirName + "\nCheck if it exists.");
                return false;
            }
            else { return true; }
        }

        //Clear contents of a file to ensure it does not contain data from a previous run:
        static void ClearFileContents(string fileName)
        {
            StreamWriter sw = new StreamWriter(fileName, false);
            sw.Close();
        }

        static string[] CreateRequestedFieldNames()
        {
            //The first 6 fields are T&C fields, the last 4 are pricing fields:
            string[] requestedFieldNames = { "RIC", "CUSIP", "ISIN", "SEDOL", "Issuer OrgID",
                                             "Currency Code",
                                             "Annualized Dividend Period Start Date",
                                             "Annualized Dividend Adjusted Gross Amount",
                                             "Balance Sheet - Enterprise Value",
                                             "Balance Sheet - Market Value" };
            return requestedFieldNames;
        }

        //Create an array for the instrument identifiers, populate it manually:
        static InstrumentIdentifier[] PopulateInstrumentIdentifiers()
        {
            //Create an array for the instrument identifiers, of size 5:
            InstrumentIdentifier[] instrumentIdentifiers = new InstrumentIdentifier[5];

            //In this example we populate it manually:
            instrumentIdentifiers[0] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Cusip,
                Identifier = "00209tab1"
            };
            instrumentIdentifiers[1] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Ric,
                Identifier = "IBM.N"
            };
            instrumentIdentifiers[2] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Isin,
                Identifier = "US4592001014"
            };
            instrumentIdentifiers[3] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Sedol,
                Identifier = "B1YW440"
            };
            //Inactive instrument:
            instrumentIdentifiers[4] = new InstrumentIdentifier
            {
                IdentifierType = IdentifierType.Cusip,
                Identifier = "438516AC0"
            };
            return instrumentIdentifiers;
        }

        static void DisplayAndLogExtractedDataFieldNames(
            string outputFile, DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            //Extract the names of the data fields, from the first row of returned data:
            StringBuilder returnedFieldNames = new StringBuilder(2048);
            ExtractionRow firstRow = extractionDataRows.First();
            foreach (string fieldName in firstRow.DynamicProperties.Keys)
            {
                returnedFieldNames.Append(fieldName);
                returnedFieldNames.Append(",");
            }
            DebugPrintAndWaitForEnter("Returned list of field names:\n" + returnedFieldNames.ToString());
            sw.WriteLine(returnedFieldNames.ToString());
            sw.Close();
        }

        static void DisplayAndLogExtractedDataFieldValues(
            string outputFile, DssCollection<ExtractionRow> extractionDataRows)
        {
            //Error handling:
            if (!extractionDataRows.Any())
            {
                DebugPrintAndWaitForEnter("No data returned ! Is it a banking holiday ?");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            //Extract the actual data values, from all rows of returned data:
            Console.WriteLine("Returned field values:");
            Console.WriteLine("======================");
            string rowValuesInString = string.Empty;
            int validDataRows = 0;
            foreach (ExtractionRow row in extractionDataRows)
            {
                IEnumerable<object> rowValues = row.DynamicProperties.Select(dp => dp.Value);
                rowValuesInString = String.Join(", ", rowValues);
                if (rowValuesInString != string.Empty)
                {
                    Console.WriteLine(rowValuesInString);
                    sw.WriteLine(rowValuesInString);
                    validDataRows++;
                }
            }

            //The number of data rows could be less than the number of instruments,
            //if data is not available for some instruments:
            DebugPrintAndWaitForEnter("\nExtraction completed." +
                "\nNumber of data rows: " + extractionDataRows.Count() +
                "\nNumber of valid (non empty) data rows: " + validDataRows +
                "\n\nOutput was also written to file.");
            sw.Close();
        }

        static void DisplayAndLogExtractionNotes(string outputFile, ExtractionResult extractionResult)
        {
            //Error handling:
            if (!extractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");
            foreach (string note in extractionResult.Notes)
            {
                Console.WriteLine(note);
                sw.WriteLine(note);
            }
            sw.Close();
            DebugPrintAndWaitForEnter("\nThese notes were also written to file");
        }

        static void DisplayAndLogAndAnalyzeExtractionNotes(string outputFile, string errorOutputFile, ExtractionResult extractionResult)
        {
            //Error handling:
            if (!extractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string inactiveMsgs = "";
            string invalidMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in extractionResult.Notes)
            {
                Console.WriteLine(notes);
                sw.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("inactive")) { inactiveMsgs = inactiveMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("invalid")) { invalidMsgs = invalidMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            sw.Close();
            Console.WriteLine("===============================================================================");

            StreamWriter swErr = new StreamWriter(errorOutputFile, true);
            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                swErr.WriteLine("\nERROR messages in notes file:\n" + errorMsgs);
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                swErr.WriteLine("WARNING messages in notes file:\n" + warningMsgs);
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (inactiveMsgs != "")
            {
                swErr.WriteLine("Inactive instruments messages in notes file:\n" + inactiveMsgs);
                Console.WriteLine("Inactive instruments messages:\n" + inactiveMsgs);
            }
            if (invalidMsgs != "")
            {
                swErr.WriteLine("Invalid instruments messages in notes file:\n" + invalidMsgs);
                Console.WriteLine("Invalid instruments messages:\n" + invalidMsgs);
            }
            if (permissionMsgs != "")
            {
                swErr.WriteLine("PERMISSION ISSUES messages in notes file:\n" + permissionMsgs);
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            swErr.Close();
            DebugPrintAndWaitForEnter("===============================================================================");
        }

        static void DisplayAndLogAndAnalyzeRawExtractionNotes(string outputFile, string errorOutputFile, RawExtractionResult rawExtractionResult)
        {
            //Error handling:
            if (!rawExtractionResult.Notes.Any())
            {
                DebugPrintAndWaitForEnter("Error: no extraction notes returned");
                return;
            }

            StreamWriter sw = new StreamWriter(outputFile, true);
            Console.WriteLine("Extraction Notes:");
            Console.WriteLine("=================");

            Boolean success = false;
            string errorMsgs = "";
            string warningMsgs = "";
            string inactiveMsgs = "";
            string invalidMsgs = "";
            string permissionMsgs = "";
            foreach (String notes in rawExtractionResult.Notes)
            {
                Console.WriteLine(notes);
                sw.WriteLine(notes);
                //The returned notes are in a single string. To analyse the contents line by line we split it:
                string[] notesLines = notes.Split(new string[] { "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string notesLine in notesLines)
                {
                    success = success || (notesLine.Contains("Processing completed successfully"));
                    if (notesLine.Contains("ERROR")) { errorMsgs = errorMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("WARNING")) { warningMsgs = warningMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("inactive")) { inactiveMsgs = inactiveMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("invalid")) { invalidMsgs = invalidMsgs + notesLine + "\n"; }
                    if (notesLine.Contains("row suppressed for lack of")) { permissionMsgs = permissionMsgs + notesLine + "\n"; }
                }
            }
            sw.Close();
            Console.WriteLine("===============================================================================");

            StreamWriter swErr = new StreamWriter(errorOutputFile, true);
            if (success) { Console.WriteLine("SUCCESS: processing completed successfully.\n"); }
            else { errorMsgs = "ERROR: processing did not complete successfully !\n" + errorMsgs; }
            if (errorMsgs != "")
            {
                swErr.WriteLine("\nERROR messages in notes file:\n" + errorMsgs);
                Console.WriteLine("ERROR messages:\n" + errorMsgs);
            }
            if (warningMsgs != "")
            {
                swErr.WriteLine("WARNING messages in notes file:\n" + warningMsgs);
                Console.WriteLine("WARNING messages:\n" + warningMsgs);
            }
            if (inactiveMsgs != "")
            {
                swErr.WriteLine("Inactive instruments messages in notes file:\n" + inactiveMsgs);
                Console.WriteLine("Inactive instruments messages:\n" + inactiveMsgs);
            }
            if (invalidMsgs != "")
            {
                swErr.WriteLine("Invalid instruments messages in notes file:\n" + invalidMsgs);
                Console.WriteLine("Invalid instruments messages:\n" + invalidMsgs);
            }
            if (permissionMsgs != "")
            {
                swErr.WriteLine("PERMISSION ISSUES messages in notes file:\n" + permissionMsgs);
                Console.WriteLine("PERMISSION ISSUES messages:\n" + permissionMsgs);
            }
            swErr.Close();
            DebugPrintAndWaitForEnter("===============================================================================");
        }
    }
}